/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   philip zhao <philip.zhao@sun.com>
 *   Seth Spitzer <sspitzer@netscape.com>
 *   Brodie Thiesfield <brofield@jellycan.com>
 *   Postbox, Inc. <contact@postbox-inc.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "prsystem.h"

#include "nsMessenger.h"

// xpcom
#include "nsIComponentManager.h"
#include "nsIServiceManager.h"
#include "nsIStringStream.h"
#include "nsReadableUtils.h"
#include "nsILocalFile.h"
#include "nsDirectoryServiceDefs.h"
#include "nsISupportsObsolete.h"
#include "nsQuickSort.h"
#include "nsAutoPtr.h"
#ifdef XP_MACOSX
#include "nsIAppleFileDecoder.h"
#include "nsILocalFileMac.h"
#endif
#include "nsNativeCharsetUtils.h"

// necko
#include "nsMimeTypes.h"
#include "nsIURL.h"
#include "nsIPrompt.h"
#include "nsIStreamListener.h"
#include "nsIStreamConverterService.h"
#include "nsNetUtil.h"
#include "nsIFileURL.h"

// rdf
#include "nsIRDFCompositeDataSource.h"
#include "nsIRDFResource.h"
#include "nsIRDFService.h"
#include "nsRDFCID.h"

// gecko
#include "nsLayoutCID.h"
#include "nsIMarkupDocumentViewer.h"
#include "nsIContentViewer.h"

// embedding
#include "nsIWebBrowserPrint.h"

/* for access to docshell */
#include "nsPIDOMWindow.h"
#include "nsIDocShell.h"
#include "nsIDocShellLoadInfo.h"
#include "nsIDocShellTreeItem.h"
#include "nsIDocShellTreeNode.h"
#include "nsIWebNavigation.h"

// mail
#include "nsIMsgMailNewsUrl.h"
#include "nsMsgUtils.h"
#include "nsMsgBaseCID.h"
#include "nsIMsgAccountManager.h"
#include "nsIMsgMailSession.h"
#include "nsIMailboxUrl.h"
#include "nsIMsgFolder.h"
#include "nsMsgFolderFlags.h"
#include "nsIMsgIncomingServer.h"

#include "nsIMsgMessageService.h"
#include "nsMsgRDFUtils.h"

#include "nsIMsgHdr.h"
#include "nsIMimeMiscStatus.h"
// compose
#include "nsMsgCompCID.h"
#include "nsMsgI18N.h"
#include "nsNativeCharsetUtils.h"

// draft/folders/sendlater/etc
#include "nsIMsgCopyService.h"
#include "nsIMsgCopyServiceListener.h"
#include "nsIUrlListener.h"

// undo
#include "nsITransaction.h"
#include "nsMsgTxn.h"

// charset conversions
#include "nsMsgMimeCID.h"
#include "nsIMimeConverter.h"

// Printing
#include "nsMsgPrintEngine.h"

// Save As
#include "nsIFilePicker.h"
#include "nsIStringBundle.h"
#include "nsIPrefService.h"
#include "nsIPrefBranch.h"
#include "nsIPrefBranch2.h"
#include "nsCExternalHandlerService.h"
#include "nsIExternalProtocolService.h"
#include "nsIMIMEService.h"
#include "nsITransfer.h"

#include "nsILinkHandler.h"

#include "nsISupportsPrimitives.h"
#include "nsIArray.h"

#include "prprf.h"

static NS_DEFINE_CID(kRDFServiceCID,  NS_RDFSERVICE_CID);
static NS_DEFINE_CID(kMsgSendLaterCID, NS_MSGSENDLATER_CID);

#define FOUR_K 4096
#define MESSENGER_SAVE_DIR_PREF_NAME "messenger.save.dir"
#define MAILNEWS_ALLOW_PLUGINS_PREF_NAME "mailnews.message_display.allow.plugins"

#define MIMETYPE_DELETED    "text/x-moz-deleted"

#define ATTACHMENT_PERMISSION 00664

//
// Convert an nsString buffer to plain text...
//
#include "nsIParser.h"
#include "nsParserCIID.h"
#include "nsICharsetConverterManager.h"
#include "nsIContentSink.h"
#include "nsIHTMLToTextSink.h"
#include "nsIHTMLContentSink.h"
#include "nsIContentSerializer.h"
#include "nsLayoutCID.h"
#include "nsIDocumentEncoder.h"
#include "nsIDOMDocumentFragment.h"
#include "nsIDOMDocument.h"
#include "nsIDOMSerializer.h"
#include "nsIFragmentContentSink.h"
#include "nsIDOMNodeList.h"
#include "nsIDOMHTMLElement.h"
#include "nsIDOMNSHTMLElement.h"
#include "nsIMsgGlobalSearch.h"
#include "nsIDOM3Node.h"

static NS_DEFINE_CID(kCParserCID, NS_PARSER_CID);
static NS_DEFINE_CID(kHTMLDocumentCID, NS_HTMLDOCUMENT_CID);

static nsresult
ConvertBufToPlainText(nsString &aConBuf)
{
  if (aConBuf.IsEmpty())
    return NS_OK;

  nsresult rv;
  nsCOMPtr<nsIParser> parser = do_CreateInstance(kCParserCID, &rv);
  if (NS_SUCCEEDED(rv) && parser)
  {
    nsCOMPtr<nsIContentSink> sink;

    sink = do_CreateInstance(NS_PLAINTEXTSINK_CONTRACTID);
    NS_ENSURE_TRUE(sink, NS_ERROR_FAILURE);

    nsCOMPtr<nsIHTMLToTextSink> textSink(do_QueryInterface(sink));
    NS_ENSURE_TRUE(textSink, NS_ERROR_FAILURE);

    nsAutoString convertedText;
    textSink->Initialize(&convertedText, 0, 72);

    parser->SetContentSink(sink);
    parser->Parse(aConBuf, 0, NS_LITERAL_CSTRING("text/html"), PR_TRUE);

    //
    // Now if we get here, we need to get from ASCII text to
    // UTF-8 format or there is a problem downstream...
    //
    if (NS_SUCCEEDED(rv))
    {
      aConBuf = convertedText;
    }
  }

  return rv;
}

nsresult ConvertAndSanitizeFileName(const char * displayName, PRUnichar ** unicodeResult, char ** result)
{
  nsCString unescapedName;

  /* we need to convert the UTF-8 fileName to platform specific character set.
     The display name is in UTF-8 because it has been escaped from JS
  */
  MsgUnescapeString(nsDependentCString(displayName), 0, unescapedName);
  NS_ConvertUTF8toUTF16 ucs2Str(unescapedName);

  nsresult rv = NS_OK;

  // replace platform specific path separator and illegale characters to avoid any confusion
  ucs2Str.ReplaceChar(FILE_PATH_SEPARATOR FILE_ILLEGAL_CHARACTERS, '-');

  if (result) {
    nsCAutoString nativeStr;
    rv =  NS_CopyUnicodeToNative(ucs2Str, nativeStr);
    *result = ToNewCString(nativeStr);
  }

  if (unicodeResult)
    *unicodeResult = ToNewUnicode(ucs2Str);

 return rv;
}

// ***************************************************
// jefft - this is a rather obscured class serves for Save Message As File,
// Save Message As Template, and Save Attachment to a file
//
class nsSaveAllAttachmentsState;

class nsSaveMsgListener : public nsIUrlListener,
                          public nsIMsgCopyServiceListener,
                          public nsIStreamListener,
                          public nsICancelable
{
public:
  nsSaveMsgListener(nsIFile *file, nsMessenger *aMessenger, nsIUrlListener *aListener);
  virtual ~nsSaveMsgListener();

  NS_DECL_ISUPPORTS

  NS_DECL_NSIURLLISTENER
  NS_DECL_NSIMSGCOPYSERVICELISTENER
  NS_DECL_NSISTREAMLISTENER
  NS_DECL_NSIREQUESTOBSERVER
  NS_DECL_NSICANCELABLE

  nsCOMPtr<nsIFile> m_file;
  nsCOMPtr<nsIOutputStream> m_outputStream;
  char *m_dataBuffer;
  nsCOMPtr<nsIChannel> m_channel;
  nsCString m_templateUri;
  nsMessenger *m_messenger; // not ref counted
  nsSaveAllAttachmentsState *m_saveAllAttachmentsState;

  // rhp: For character set handling
  PRBool        m_doCharsetConversion;
  nsString      m_charset;
  enum {
    eUnknown,
    ePlainText,
    eHTML
  }             m_outputFormat;
  nsString      m_msgBuffer;

  nsCString     m_contentType;    // used only when saving attachment

  nsCOMPtr<nsITransfer> mTransfer;
  nsCOMPtr<nsIUrlListener> mListener;
  nsCOMPtr<nsIURI> mListenerUri;
  PRInt64 mProgress;
  PRInt64 mMaxProgress;
  PRBool  mCanceled;
  PRBool  mInitialized;
  PRBool  mUrlHasStopped;
  PRBool  mRequestHasStopped;
  nsresult InitializeDownload(nsIRequest * aRequest, PRUint32 aBytesDownloaded);
};

class nsSaveAllAttachmentsState
{
public:
  nsSaveAllAttachmentsState(PRUint32 count,
                            const char **contentTypeArray,
                            const char **urlArray,
                            const char **displayNameArray,
                            const char **messageUriArray,
                            const char *directoryName,
                            PRBool detachingAttachments);
  virtual ~nsSaveAllAttachmentsState();

  PRUint32 m_count;
  PRUint32 m_curIndex;
  char* m_directoryName;
  char** m_contentTypeArray;
  char** m_urlArray;
  char** m_displayNameArray;
  char** m_messageUriArray;
  PRBool m_detachingAttachments;

  // if detaching, do without warning? Will create unique files instead of
  //  prompting if duplicate files exist.
  PRBool m_withoutWarning;
  nsCStringArray m_savedFiles; // if detaching first, remember where we saved to.
};

//
// nsMessenger
//
nsMessenger::nsMessenger()
{
  mScriptObject = nsnull;
  mCurHistoryPos = -2; // first message selected goes at position 0.
  //  InitializeFolderRoot();
}

nsMessenger::~nsMessenger()
{
  // Release search context.
  mSearchContext = nsnull;
}


NS_IMPL_ISUPPORTS4(nsMessenger, nsIMessenger, nsIObserver, nsISupportsWeakReference, nsIFolderListener)

NS_IMETHODIMP nsMessenger::SetWindow(nsIDOMWindowInternal *aWin, nsIMsgWindow *aMsgWindow)
{
  nsresult rv;

  nsCOMPtr<nsIPrefBranch2> pbi = do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIMsgMailSession> mailSession =
    do_GetService(NS_MSGMAILSESSION_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  if (aWin)
  {
    mMsgWindow = aMsgWindow;
    mWindow = aWin;

    rv = mailSession->AddFolderListener(this, nsIFolderListener::removed);
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsPIDOMWindow> win( do_QueryInterface(aWin) );
    NS_ENSURE_TRUE(win, NS_ERROR_FAILURE);
    nsIDocShell *docShell = win->GetDocShell();

    nsCOMPtr<nsIDocShellTreeItem> docShellAsItem(do_QueryInterface(docShell));
    NS_ENSURE_TRUE(docShellAsItem, NS_ERROR_FAILURE);

    nsCOMPtr<nsIDocShellTreeItem> rootDocShellAsItem;
    docShellAsItem->GetSameTypeRootTreeItem(getter_AddRefs(rootDocShellAsItem));

    nsCOMPtr<nsIDocShellTreeNode> rootDocShellAsNode(do_QueryInterface(rootDocShellAsItem));
    if (rootDocShellAsNode)
    {
      nsCOMPtr<nsIDocShellTreeItem> childAsItem;
      rv = rootDocShellAsNode->FindChildWithName(NS_LITERAL_STRING("messagepane").get(),
                                                 PR_TRUE, PR_FALSE, nsnull, nsnull, getter_AddRefs(childAsItem));

      mDocShell = do_QueryInterface(childAsItem);
      if (NS_SUCCEEDED(rv) && mDocShell) {
        mCurrentDisplayCharset = ""; // Important! Clear out mCurrentDisplayCharset so we reset a default charset on mDocshell the next time we try to load something into it.

        if (aMsgWindow)
        {
          aMsgWindow->GetTransactionManager(getter_AddRefs(mTxnMgr));
          // Add pref observer
          pbi->AddObserver(MAILNEWS_ALLOW_PLUGINS_PREF_NAME, this, PR_TRUE);
          SetDisplayProperties();
        }
      }
    }

    // we don't always have a message pane, like in the addressbook
    // so if we don't have a docshell, use the one for the xul window.
    // we do this so OpenURL() will work.
    if (!mDocShell)
      mDocShell = docShell;
  } // if aWin
  else
  {
    // it isn't an error to pass in null for aWin, in fact it means we are shutting
    // down and we should start cleaning things up...
    // Remove pref observer
    pbi->RemoveObserver(MAILNEWS_ALLOW_PLUGINS_PREF_NAME, this);

    // Remove the folder listener if we added it, i.e. if mWindow is non-null
    if (mWindow)
    {
      rv = mailSession->RemoveFolderListener(this);
      NS_ENSURE_SUCCESS(rv, rv);
    }

    mWindow = nsnull;
  }

  return NS_OK;
}

NS_IMETHODIMP nsMessenger::SetDisplayCharset(const nsACString& aCharset)
{
  // libmime always converts to UTF-8 (both HTML and XML)
  if (mDocShell)
  {
    nsCOMPtr<nsIContentViewer> cv;
    mDocShell->GetContentViewer(getter_AddRefs(cv));
    if (cv)
    {
      nsCOMPtr<nsIMarkupDocumentViewer> muDV = do_QueryInterface(cv);
      if (muDV)
      {
        muDV->SetHintCharacterSet(aCharset);
        muDV->SetHintCharacterSetSource(kCharsetFromMetaTag);
      }

      mCurrentDisplayCharset = aCharset;
    }
  }

  return NS_OK;
}

nsresult
nsMessenger::SetDisplayProperties()
{
  // For now, the only property we will set is allowPlugins but we might do more in the future...

  nsresult rv;

  if (!mDocShell)
    return NS_ERROR_FAILURE;

  PRBool allowPlugins = PR_FALSE;

  nsCOMPtr<nsIPrefBranch> prefBranch = do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
  if (NS_SUCCEEDED(rv))
    prefBranch->GetBoolPref(MAILNEWS_ALLOW_PLUGINS_PREF_NAME, &allowPlugins);

  return mDocShell->SetAllowPlugins(allowPlugins);
}

NS_IMETHODIMP
nsMessenger::Observe(nsISupports *aSubject, const char *aTopic, const PRUnichar *aData)
{
  if (!strcmp(aTopic, NS_PREFBRANCH_PREFCHANGE_TOPIC_ID))
  {
    nsDependentString prefName(aData);
    if (prefName.EqualsLiteral(MAILNEWS_ALLOW_PLUGINS_PREF_NAME))
      SetDisplayProperties();
  }

  return NS_OK;
}

nsresult
nsMessenger::PromptIfFileExists(nsILocalFile *file)
{
  nsresult rv = NS_ERROR_FAILURE;
  PRBool exists;
  file->Exists(&exists);
  if (exists)
  {
    nsCOMPtr<nsIPrompt> dialog(do_GetInterface(mDocShell));
    if (!dialog) return rv;
    nsAutoString path;
    PRBool dialogResult = PR_FALSE;
    nsString errorMessage;

    file->GetPath(path);
    const PRUnichar *pathFormatStrings[] = { path.get() };

    if (!mStringBundle)
    {
      rv = InitStringBundle();
      NS_ENSURE_SUCCESS(rv, rv);
    }
    rv = mStringBundle->FormatStringFromName(NS_LITERAL_STRING("fileExists").get(),
                                             pathFormatStrings, 1,
                                             getter_Copies(errorMessage));
    NS_ENSURE_SUCCESS(rv, rv);
    rv = dialog->Confirm(nsnull, errorMessage.get(), &dialogResult);
    NS_ENSURE_SUCCESS(rv, rv);

    if (dialogResult)
    {
      return NS_OK; // user says okay to replace
    }
    else
    {
      // if we don't re-init the path for redisplay the picker will
      // show the full path, not just the file name
      nsCOMPtr<nsILocalFile> currentFile = do_CreateInstance("@mozilla.org/file/local;1");
      if (!currentFile) return NS_ERROR_FAILURE;

      rv = currentFile->InitWithPath(path);
      NS_ENSURE_SUCCESS(rv, rv);

      nsAutoString leafName;
      currentFile->GetLeafName(leafName);
      if (!leafName.IsEmpty())
        path.Assign(leafName); // path should be a copy of leafName

      nsCOMPtr<nsIFilePicker> filePicker =
        do_CreateInstance("@mozilla.org/filepicker;1", &rv);
      NS_ENSURE_SUCCESS(rv, rv);
      nsString saveAttachmentStr;
      GetString(NS_LITERAL_STRING("SaveAttachment"), saveAttachmentStr);
      filePicker->Init(mWindow,
                       saveAttachmentStr,
                       nsIFilePicker::modeSave);
      filePicker->SetDefaultString(path);
      filePicker->AppendFilters(nsIFilePicker::filterAll);

      nsCOMPtr <nsILocalFile> lastSaveDir;
      rv = GetLastSaveDirectory(getter_AddRefs(lastSaveDir));
      if (NS_SUCCEEDED(rv) && lastSaveDir) {
        filePicker->SetDisplayDirectory(lastSaveDir);
      }

      PRInt16 dialogReturn;
      rv = filePicker->Show(&dialogReturn);
      if (NS_FAILED(rv) || dialogReturn == nsIFilePicker::returnCancel) {
        // XXX todo
        // don't overload the return value like this
        // change this function to have an out boolean
        // that we check to see if the user cancelled
        return NS_ERROR_FAILURE;
      }

      nsCOMPtr<nsILocalFile> localFile;

      rv = filePicker->GetFile(getter_AddRefs(localFile));
      NS_ENSURE_SUCCESS(rv, rv);

      rv = SetLastSaveDirectory(localFile);
      NS_ENSURE_SUCCESS(rv,rv);

      // reset the file to point to the new path
      return file->InitWithFile(localFile);
    }
  }
  else
  {
    return NS_OK;
  }
  return rv;
}

void nsMessenger::AddMsgUrlToNavigateHistory(const nsACString& aURL)
{
  // mNavigatingToUri is set to a url if we're already doing a back/forward,
  // in which case we don't want to add the url to the history list.
  // Or if the entry at the cur history pos is the same as what we're loading, don't
  // add it to the list.
  if (!mNavigatingToUri.Equals(aURL) && (mCurHistoryPos < 0 || !mLoadedMsgHistory[mCurHistoryPos]->Equals(aURL)))
  {
    mNavigatingToUri = aURL;
    nsCString curLoadedFolderUri;
    nsCOMPtr <nsIMsgFolder> curLoadedFolder;

    mMsgWindow->GetOpenFolder(getter_AddRefs(curLoadedFolder));
    // for virtual folders, we want to select the right folder,
    // which isn't the same as the folder specified in the msg uri.
    // So add the uri for the currently loaded folder to the history list.
    if (curLoadedFolder)
      curLoadedFolder->GetURI(curLoadedFolderUri);

    mLoadedMsgHistory.InsertCStringAt(mNavigatingToUri, mCurHistoryPos++ + 2);
    mLoadedMsgHistory.InsertCStringAt(curLoadedFolderUri, mCurHistoryPos++ + 2);
    // we may want to prune this history if it gets large, but I think it's
    // more interesting to prune the back and forward menu.
  }
}

/* XXX - aMessageURIs really should really be a utf8 string enumerator, but you can't reset that enumerator
 * and nsMessenger needs to peek at the first message URI before passing the enumerator out to 
 * the display emitter....
 */
NS_IMETHODIMP nsMessenger::Display(nsIArray * aMessageURIs, nsIMsgDBHdr * aPrimaryMsgHdr, PRBool aCustomMessages, PRUint32 aDisplayOutputFlags)
{
  nsresult rv;
  NS_ENSURE_ARG_POINTER(aMessageURIs);

  // This is to setup the display DocShell as UTF-8 capable...
  SetDisplayCharset(NS_LITERAL_CSTRING("UTF-8"));
  PRUint32 numMessages = 0;
  aMessageURIs->GetLength(&numMessages);
  nsCString msgURI;
  // extract the first message in the list and use that to set mLastDisplayURI and AddMsgUrlToNavigateHistory
  nsCOMPtr<nsISupportsCString> str;
  if (numMessages)
  {
    aMessageURIs->QueryElementAt(0, NS_GET_IID(nsISupportsCString), getter_AddRefs(str));
    str->GetData(msgURI);
  }

  // For regular messages, aURI is a resource URI (mailbox-message, imap-message, etc. which can be passed direcly to
  // our display emitter. If we were given a string that begins with file or contains the string 'type=application/x-message-display' then
  // we are opening a custom message (either a .eml file or a message attachment that's a .eml file). For this later case,
  // we have some work to do.
  if (aCustomMessages)
  {
    for (PRUint32 index = 0; index < numMessages; index++)
    {
      nsCOMPtr<nsISupportsCString> strSupports;
      aMessageURIs->QueryElementAt(index, NS_GET_IID(nsISupportsCString), getter_AddRefs(strSupports));
      nsCString str;
      strSupports->GetData(str);

      // we have to do some string cleanup to do first
      NS_ConvertASCIItoUTF16 uriString(str);
      // Cleanup the empty spaces that might be on each end.
      uriString.Trim(" ");
      // Eliminate embedded newlines, which single-line text fields now allow:
      uriString.StripChars("\r\n");
      NS_ENSURE_TRUE(!uriString.IsEmpty(), NS_ERROR_FAILURE);

      PRBool loadingFromFile = PR_FALSE;
      PRInt64 fileSize;
      // turn file urls into mailbox urls
      if (StringBeginsWith(uriString, NS_LITERAL_STRING("file:")))
      {
        nsCOMPtr<nsIURI> fileUri;
        rv = NS_NewURI(getter_AddRefs(fileUri), uriString);
        NS_ENSURE_SUCCESS(rv, rv);
        nsCOMPtr <nsIFileURL> fileUrl = do_QueryInterface(fileUri, &rv);
        NS_ENSURE_SUCCESS(rv, rv);
        nsCOMPtr <nsIFile> file;
        rv = fileUrl->GetFile(getter_AddRefs(file));
        NS_ENSURE_SUCCESS(rv, rv);
        file->GetFileSize(&fileSize);
        uriString.ReplaceSubstring(NS_LITERAL_STRING("file:"), NS_LITERAL_STRING("mailbox:"));
        uriString.Append(NS_LITERAL_STRING("&number=0"));
        loadingFromFile = PR_TRUE;
      }

      nsCOMPtr<nsIURI> uri;
      rv = NS_NewURI(getter_AddRefs(uri), uriString);
      NS_ENSURE_SUCCESS(rv, rv);
      nsCOMPtr<nsIMsgMailNewsUrl> msgurl = do_QueryInterface(uri, &rv);
      NS_ENSURE_SUCCESS(rv, rv);
      msgurl->SetMsgWindow(mMsgWindow);
      if (loadingFromFile)
      {
        nsCOMPtr <nsIMailboxUrl> mailboxUrl = do_QueryInterface(msgurl, &rv);
        mailboxUrl->SetMessageSize((PRUint32) fileSize);
      }

      nsCOMPtr <nsIMsgHeaderSink> headerSink;
      // need to tell the header sink to capture some headers to create a fake db header
      // so we can do reply to a .eml file or a rfc822 msg attachment.
      mMsgWindow->GetMsgHeaderSink(getter_AddRefs(headerSink));
      if (headerSink)
      {
        nsCOMPtr <nsIMsgDBHdr> dummyHeader;
        headerSink->GetDummyMsgHeader(getter_AddRefs(dummyHeader));
        if (dummyHeader && loadingFromFile)
          dummyHeader->SetMessageSize((PRUint32) fileSize);
      }
      nsCString uriSpec;
      uri->GetSpec(uriSpec);
      strSupports->SetData(uriSpec); // replace the value of our string in the array

      // conversation handling
      if (!mDisplayEmitter)
        mDisplayEmitter = do_CreateInstance("@mozilla.org/messenger/displayemitter;1");
      nsCOMPtr<nsIWebNavigation> webNav(do_QueryInterface(mDocShell));
      webNav->Stop(nsIWebNavigation::STOP_ALL);
      if (!mDisplayEmitter)
        return NS_ERROR_UNEXPECTED;
      mDisplayEmitter->Display(aMessageURIs, nsnull /* aPrimaryMsgHdr */, aCustomMessages, mDocShell, mMsgWindow, aDisplayOutputFlags);
    } // for loop
    return NS_OK;
  } // if aCustomMessage

  // conversation handling
  if (!mDisplayEmitter)
    mDisplayEmitter = do_CreateInstance("@mozilla.org/messenger/displayemitter;1");
  nsCOMPtr<nsIWebNavigation> webNav(do_QueryInterface(mDocShell));
  webNav->Stop(nsIWebNavigation::STOP_ALL);
  if (!mDisplayEmitter)
    return NS_ERROR_UNEXPECTED;
  mDisplayEmitter->Display(aMessageURIs, aPrimaryMsgHdr, aCustomMessages, mDocShell, mMsgWindow, aDisplayOutputFlags);

  AddMsgUrlToNavigateHistory(msgURI);
  mLastDisplayURI = msgURI; // remember the last uri we displayed....
  return NS_OK;
}

NS_IMETHODIMP nsMessenger::OpenURL(const nsACString& aURL)
{
  nsresult rv;
  // This is to setup Athe display DocShell as UTF-8 capable...
  SetDisplayCharset(NS_LITERAL_CSTRING("UTF-8"));

  NS_ConvertASCIItoUTF16 uriString(aURL);
  // Cleanup the empty spaces that might be on each end.
  uriString.Trim(" ");
  // Eliminate embedded newlines, which single-line text fields now allow:
  uriString.StripChars("\r\n");
  NS_ENSURE_TRUE(!uriString.IsEmpty(), NS_ERROR_FAILURE);

  nsCOMPtr<nsIURI> uri;
  rv = NS_NewURI(getter_AddRefs(uri), uriString);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIMsgMailNewsUrl> msgurl = do_QueryInterface(uri);
  if (msgurl)
    msgurl->SetMsgWindow(mMsgWindow);
    
  nsCOMPtr<nsIWebNavigation> webNav(do_QueryInterface(mDocShell));
  return webNav->LoadURI(NS_ConvertASCIItoUTF16(aURL).get(),   // URI string
                       nsIWebNavigation::LOAD_FLAGS_IS_LINK, // Load flags
                       nsnull,                               // Referring URI
                       nsnull,                               // Post stream
                       nsnull);                              // Extra headers
}


NS_IMETHODIMP nsMessenger::LaunchExternalURL(const nsACString& aURL)
{
  nsresult rv;

  nsCOMPtr<nsIURI> uri;
  rv = NS_NewURI(getter_AddRefs(uri), PromiseFlatCString(aURL).get());
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIExternalProtocolService> extProtService = do_GetService(NS_EXTERNALPROTOCOLSERVICE_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv,rv);
  return extProtService->LoadUrl(uri);
}

NS_IMETHODIMP
nsMessenger::LoadURL(nsIDOMWindowInternal *aWin, const nsACString& aURL)
{
  nsresult rv;

  SetDisplayCharset(NS_LITERAL_CSTRING("UTF-8"));

  NS_ConvertASCIItoUTF16 uriString(aURL);
  // Cleanup the empty spaces that might be on each end.
  uriString.Trim(" ");
  // Eliminate embedded newlines, which single-line text fields now allow:
  uriString.StripChars("\r\n");
  NS_ENSURE_TRUE(!uriString.IsEmpty(), NS_ERROR_FAILURE);

  PRBool loadingFromFile = PR_FALSE;
  PRBool getDummyMsgHdr = PR_FALSE;
  PRInt64 fileSize;

  if (StringBeginsWith(uriString, NS_LITERAL_STRING("file:")))
  {
    nsCOMPtr<nsIURI> fileUri;
    rv = NS_NewURI(getter_AddRefs(fileUri), uriString);
    NS_ENSURE_SUCCESS(rv, rv);
    nsCOMPtr <nsIFileURL> fileUrl = do_QueryInterface(fileUri, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    nsCOMPtr <nsIFile> file;
    rv = fileUrl->GetFile(getter_AddRefs(file));
    NS_ENSURE_SUCCESS(rv, rv);
    file->GetFileSize(&fileSize);
    uriString.ReplaceSubstring(NS_LITERAL_STRING("file:"), NS_LITERAL_STRING("mailbox:"));
    uriString.Append(NS_LITERAL_STRING("&number=0"));
    loadingFromFile = PR_TRUE;
    getDummyMsgHdr = PR_TRUE;
  }
  else if (uriString.Find("type=application/x-message-display") >= 0)
    getDummyMsgHdr = PR_TRUE;

  nsCOMPtr<nsIURI> uri;
  rv = NS_NewURI(getter_AddRefs(uri), uriString);
  NS_ENSURE_SUCCESS(rv, rv);

  NS_ENSURE_TRUE(mDocShell, NS_ERROR_FAILURE);
  nsCOMPtr<nsIMsgMailNewsUrl> msgurl = do_QueryInterface(uri);
  if (msgurl)
  {
    msgurl->SetMsgWindow(mMsgWindow);
    if (loadingFromFile || getDummyMsgHdr)
    {
      if (loadingFromFile)
      {
        nsCOMPtr <nsIMailboxUrl> mailboxUrl = do_QueryInterface(msgurl, &rv);
        mailboxUrl->SetMessageSize((PRUint32) fileSize);
      }
      if (getDummyMsgHdr)
      {
        nsCOMPtr <nsIMsgHeaderSink> headerSink;
         // need to tell the header sink to capture some headers to create a fake db header
         // so we can do reply to a .eml file or a rfc822 msg attachment.
        mMsgWindow->GetMsgHeaderSink(getter_AddRefs(headerSink));
        if (headerSink)
        {
          nsCOMPtr <nsIMsgDBHdr> dummyHeader;
          headerSink->GetDummyMsgHeader(getter_AddRefs(dummyHeader));
          if (dummyHeader && loadingFromFile)
            dummyHeader->SetMessageSize((PRUint32) fileSize);
        }
      }
    }
  }

  nsCOMPtr<nsIDocShellLoadInfo> loadInfo;
  rv = mDocShell->CreateLoadInfo(getter_AddRefs(loadInfo));
  NS_ENSURE_SUCCESS(rv, rv);
  loadInfo->SetLoadType(nsIDocShellLoadInfo::loadNormal);
  AddMsgUrlToNavigateHistory(aURL);
  mNavigatingToUri.Truncate();
  return mDocShell->LoadURI(uri, loadInfo, 0, PR_TRUE);
}

NS_IMETHODIMP nsMessenger::SaveAttachmentToFile(nsIFile *aFile,
                                                const nsACString &aURL,
                                                const nsACString &aMessageUri,
                                                const nsACString &aContentType,
                                                nsIUrlListener *aListener)
{
  return SaveAttachment(aFile, aURL, aMessageUri, aContentType, nsnull, aListener);
}

NS_IMETHODIMP
nsMessenger::DetachAttachmentsWOPrompts(nsIFile* aDestFolder,
                                        PRUint32 aCount,
                                        const char **aContentTypeArray,
                                        const char **aUrlArray,
                                        const char **aDisplayNameArray,
                                        const char **aMessageUriArray,
                                        nsIUrlListener *aListener)
{
  nsSaveAllAttachmentsState *saveState;
  nsCOMPtr<nsIFile> clone;
  nsresult rv = aDestFolder->Clone(getter_AddRefs(clone));
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsILocalFile> attachmentDestination(do_QueryInterface(clone, &rv));
  NS_ENSURE_SUCCESS(rv, rv);

  nsCAutoString path;
  rv = attachmentDestination->GetNativePath(path);
  NS_ENSURE_SUCCESS(rv, rv);

  nsAutoString unescapedFileName;
  ConvertAndSanitizeFileName(aDisplayNameArray[0], getter_Copies(unescapedFileName), nsnull);
  rv = attachmentDestination->Append(unescapedFileName);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = attachmentDestination->CreateUnique(nsIFile::NORMAL_FILE_TYPE, ATTACHMENT_PERMISSION);
  NS_ENSURE_SUCCESS(rv, rv);

  saveState = new nsSaveAllAttachmentsState(aCount,
                                            aContentTypeArray,
                                            aUrlArray,
                                            aDisplayNameArray,
                                            aMessageUriArray,
                                            path.get(),
                                            PR_TRUE);

  // This method is used in filters, where we don't want to warn
  saveState->m_withoutWarning = PR_TRUE;
  rv = SaveAttachment(attachmentDestination,
                      nsDependentCString(aUrlArray[0]),
                      nsDependentCString(aMessageUriArray[0]),
                      nsDependentCString(aContentTypeArray[0]),
                      (void *)saveState,
                      aListener);
  return rv;
}

nsresult nsMessenger::SaveAttachment(nsIFile *aFile,
                                     const nsACString &aURL,
                                     const nsACString &aMessageUri,
                                     const nsACString &aContentType,
                                     void *closure,
                                     nsIUrlListener *aListener)
{
  nsCOMPtr<nsIMsgMessageService> messageService;
  nsSaveAllAttachmentsState *saveState= (nsSaveAllAttachmentsState*) closure;
  nsCOMPtr<nsIMsgMessageFetchPartService> fetchService;
  nsCAutoString urlString;
  nsCOMPtr<nsIURI> URL;
  nsCAutoString fullMessageUri(aMessageUri);

  // This instance will be held onto by the listeners, and will be released once 
  // the transfer has been completed.
  nsRefPtr<nsSaveMsgListener> saveListener(new nsSaveMsgListener(aFile, this, aListener));
  if (!saveListener)
    return NS_ERROR_OUT_OF_MEMORY;

  saveListener->m_contentType = aContentType;
  if (saveState)
  {
    saveListener->m_saveAllAttachmentsState = saveState;
    if (saveState->m_detachingAttachments)
    {
      nsCOMPtr<nsIURI> outputURI;
      nsresult rv = NS_NewFileURI(getter_AddRefs(outputURI), aFile);
      NS_ENSURE_SUCCESS(rv, rv);
      nsCAutoString fileUriSpec;
      outputURI->GetSpec(fileUriSpec);
      saveState->m_savedFiles.AppendCString(fileUriSpec);
    }
  }

  urlString = aURL;
  // strip out ?type=application/x-message-display because it confuses libmime

  PRInt32 typeIndex = urlString.Find("?type=application/x-message-display");
  if (typeIndex != kNotFound)
  {
    urlString.Cut(typeIndex, sizeof("?type=application/x-message-display") - 1);
    // we also need to replace the next '&' with '?'
    PRInt32 firstPartIndex = urlString.FindChar('&');
    if (firstPartIndex != kNotFound)
      urlString.SetCharAt('?', firstPartIndex);
  }

  urlString.ReplaceSubstring("/;section", "?section");
  nsresult rv = CreateStartupUrl(urlString.get(), getter_AddRefs(URL));

  if (NS_SUCCEEDED(rv))
  {
    // set the message Uri on the url
    nsCOMPtr<nsIMsgMessageUrl> messageUrl (do_QueryInterface(URL));
    if (messageUrl)
      messageUrl->SetUri(PromiseFlatCString(aMessageUri).get());
    rv = GetMessageServiceFromURI(aMessageUri, getter_AddRefs(messageService));
    if (NS_SUCCEEDED(rv))
    {
      fetchService = do_QueryInterface(messageService);
      // if the message service has a fetch part service then we know we can fetch mime parts...
      if (fetchService)
      {
        nsCString mimePart;
        PRInt32 sectionPos = urlString.Find("?section");
        urlString.Right(mimePart, urlString.Length() - sectionPos);
        fullMessageUri.Append(mimePart);
      }

      nsCOMPtr<nsIStreamListener> convertedListener;
      saveListener->QueryInterface(NS_GET_IID(nsIStreamListener),
                                 getter_AddRefs(convertedListener));

#ifndef XP_MACOSX
      // if the content type is bin hex we are going to do a hokey hack and make sure we decode the bin hex
      // when saving an attachment to disk..
      if (aContentType.LowerCaseEqualsLiteral(APPLICATION_BINHEX))
      {
        nsCOMPtr<nsIStreamListener> listener (do_QueryInterface(convertedListener));
        nsCOMPtr<nsIStreamConverterService> streamConverterService = do_GetService("@mozilla.org/streamConverters;1", &rv);
        nsCOMPtr<nsISupports> channelSupport = do_QueryInterface(saveListener->m_channel);

        rv = streamConverterService->AsyncConvertData(APPLICATION_BINHEX,
                                                      "*/*",
                                                      listener,
                                                      channelSupport,
                                                      getter_AddRefs(convertedListener));
      }
#endif
      if (fetchService)
        rv = fetchService->FetchMimePart(URL, fullMessageUri.get(), convertedListener, mMsgWindow, saveListener, nsnull);
      else
        rv = messageService->DisplayMessage(fullMessageUri.get(), convertedListener, mMsgWindow, nsnull, nsnull, nsnull);
    } // if we got a message service
  } // if we created a url

  if (NS_FAILED(rv))
    Alert("saveAttachmentFailed");

  return rv;
}

NS_IMETHODIMP
nsMessenger::OpenAttachment(const nsACString& aContentType, const nsACString& aURL,
                            const nsACString& aDisplayName, const nsACString& aMessageUri, PRBool aIsExternalAttachment)
{
  nsresult rv = NS_OK;

  // open external attachments inside our message pane which in turn should trigger the
  // helper app dialog...
  if (aIsExternalAttachment)
    rv = OpenURL(aURL);
  else
  {
    nsCOMPtr <nsIMsgMessageService> messageService;
    rv = GetMessageServiceFromURI(aMessageUri, getter_AddRefs(messageService));
    if (messageService)
      rv = messageService->OpenAttachment(PromiseFlatCString(aContentType).get(), PromiseFlatCString(aDisplayName).get(),
                                          PromiseFlatCString(aURL).get(), PromiseFlatCString(aMessageUri).get(),
                                          mDocShell, mMsgWindow, nsnull);
  }

  return rv;
}

NS_IMETHODIMP
nsMessenger::SaveAttachmentToFolder(const nsACString& contentType, const nsACString& url, const nsACString& displayName,
                                    const nsACString& messageUri, nsILocalFile * aDestFolder, nsIUrlListener * aUrlListener, nsILocalFile ** aOutFile)
{
  NS_ENSURE_ARG_POINTER(aDestFolder);
  nsresult rv;

  nsCOMPtr<nsIFile> clone;
  rv = aDestFolder->Clone(getter_AddRefs(clone));
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsILocalFile> attachmentDestination = do_QueryInterface(clone);

  nsCString unescapedFileName;
  rv = ConvertAndSanitizeFileName(PromiseFlatCString(displayName).get(), nsnull, getter_Copies(unescapedFileName));
  NS_ENSURE_SUCCESS(rv, rv);

  rv = attachmentDestination->AppendNative(unescapedFileName);
  NS_ENSURE_SUCCESS(rv, rv);

  // make sure the file is unique so we don't overwrite an existing file
  rv = attachmentDestination->CreateUnique(nsIFile::NORMAL_FILE_TYPE, ATTACHMENT_PERMISSION);

  rv = SaveAttachment(attachmentDestination, url, messageUri, contentType, nsnull, aUrlListener);
  attachmentDestination.swap(*aOutFile);
  return rv;
}

NS_IMETHODIMP
nsMessenger::SaveAttachment(const nsACString& aContentType, const nsACString& aURL,
                            const nsACString& aDisplayName, const nsACString& aMessageUri, PRBool aIsExternalAttachment)
{
  // open external attachments inside our message pane which in turn should trigger the
  // helper app dialog...
  if (aIsExternalAttachment)
    return OpenURL(aURL);

  nsresult rv = NS_ERROR_OUT_OF_MEMORY;
  nsCOMPtr<nsIFilePicker> filePicker =
      do_CreateInstance("@mozilla.org/filepicker;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  PRInt16 dialogResult;
  nsCOMPtr<nsILocalFile> localFile;
  nsCOMPtr<nsILocalFile> lastSaveDir;
  nsCString filePath;
  nsString saveAttachmentStr;
  nsString defaultDisplayString;
  rv = ConvertAndSanitizeFileName(PromiseFlatCString(aDisplayName).get(), getter_Copies(defaultDisplayString), nsnull);
  NS_ENSURE_SUCCESS(rv, rv);

  GetString(NS_LITERAL_STRING("SaveAttachment"), saveAttachmentStr);
  filePicker->Init(mWindow, saveAttachmentStr,
                   nsIFilePicker::modeSave);
  filePicker->SetDefaultString(defaultDisplayString);
  filePicker->AppendFilters(nsIFilePicker::filterAll);

  rv = GetLastSaveDirectory(getter_AddRefs(lastSaveDir));
  if (NS_SUCCEEDED(rv) && lastSaveDir)
    filePicker->SetDisplayDirectory(lastSaveDir);

  rv = filePicker->Show(&dialogResult);
  if (NS_FAILED(rv) || dialogResult == nsIFilePicker::returnCancel)
    return rv;

  rv = filePicker->GetFile(getter_AddRefs(localFile));
  NS_ENSURE_SUCCESS(rv, rv);

  SetLastSaveDirectory(localFile);

  rv = SaveAttachment(localFile, aURL, aMessageUri, aContentType, nsnull, nsnull);
  return rv;
}


NS_IMETHODIMP
nsMessenger::SaveAllAttachments(PRUint32 count,
                                const char **contentTypeArray,
                                const char **urlArray,
                                const char **displayNameArray,
                                const char **messageUriArray)
{
  if (!count)
    return NS_ERROR_INVALID_ARG;
  return SaveAllAttachments(count, contentTypeArray, urlArray, displayNameArray, messageUriArray, PR_FALSE);
}

nsresult
nsMessenger::SaveAllAttachments(PRUint32 count,
                                const char **contentTypeArray,
                                const char **urlArray,
                                const char **displayNameArray,
                                const char **messageUriArray,
                                PRBool detaching)
{
  nsresult rv = NS_ERROR_OUT_OF_MEMORY;
  nsCOMPtr<nsIFilePicker> filePicker =
    do_CreateInstance("@mozilla.org/filepicker;1", &rv);
  nsCOMPtr<nsILocalFile> localFile;
  nsCOMPtr<nsILocalFile> lastSaveDir;
  nsCString dirName;
  nsSaveAllAttachmentsState *saveState = nsnull;
  PRInt16 dialogResult;
  nsString saveAttachmentStr;

  NS_ENSURE_SUCCESS(rv, rv);
  GetString(NS_LITERAL_STRING("SaveAllAttachments"), saveAttachmentStr);
  filePicker->Init(mWindow,
                   saveAttachmentStr,
                   nsIFilePicker::modeGetFolder);

  rv = GetLastSaveDirectory(getter_AddRefs(lastSaveDir));
  if (NS_SUCCEEDED(rv) && lastSaveDir)
    filePicker->SetDisplayDirectory(lastSaveDir);

  rv = filePicker->Show(&dialogResult);
  if (NS_FAILED(rv) || dialogResult == nsIFilePicker::returnCancel)
    return rv;

  rv = filePicker->GetFile(getter_AddRefs(localFile));
  NS_ENSURE_SUCCESS(rv, rv);

  rv = SetLastSaveDirectory(localFile);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = localFile->GetNativePath(dirName);
  NS_ENSURE_SUCCESS(rv, rv);

  saveState = new nsSaveAllAttachmentsState(count,
                                            contentTypeArray,
                                            urlArray,
                                            displayNameArray,
                                            messageUriArray,
                                            dirName.get(), detaching);
  nsCString unescapedName;
  rv = ConvertAndSanitizeFileName(displayNameArray[0], nsnull, getter_Copies(unescapedName));
  NS_ENSURE_SUCCESS(rv, rv);

  localFile->AppendNative(unescapedName);
  rv = PromptIfFileExists(localFile);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = SaveAttachment(localFile, nsDependentCString(urlArray[0]), nsDependentCString(messageUriArray[0]),
                      nsDependentCString(contentTypeArray[0]), (void *)saveState, nsnull);
  return rv;
}

enum MESSENGER_SAVEAS_FILE_TYPE
{
 EML_FILE_TYPE =  0,
 HTML_FILE_TYPE = 1,
 TEXT_FILE_TYPE = 2,
 ANY_FILE_TYPE =  3
};
#define HTML_FILE_EXTENSION ".htm"
#define HTML_FILE_EXTENSION2 ".html"
#define TEXT_FILE_EXTENSION ".txt"
#define EML_FILE_EXTENSION  ".eml"

NS_IMETHODIMP
nsMessenger::SaveAs(const nsACString& aURI, PRBool aAsFile, nsIMsgIdentity *aIdentity, const nsAString& aMsgFilename)
{
  nsCOMPtr<nsIMsgMessageService> messageService;
  nsCOMPtr<nsIUrlListener> urlListener;
  nsSaveMsgListener *saveListener = nsnull;
  nsCOMPtr<nsIURI> url;
  nsCOMPtr<nsIStreamListener> convertedListener;
  PRInt32 saveAsFileType = EML_FILE_TYPE;

  nsresult rv = GetMessageServiceFromURI(aURI, getter_AddRefs(messageService));
  if (NS_FAILED(rv))
    goto done;

  if (aAsFile)
  {
    nsCOMPtr<nsIFilePicker> filePicker = do_CreateInstance("@mozilla.org/filepicker;1", &rv);
    if (NS_FAILED(rv))
      goto done;
    nsString saveMailAsStr;
    GetString(NS_LITERAL_STRING("SaveMailAs"), saveMailAsStr);
    filePicker->Init(mWindow, saveMailAsStr,
                     nsIFilePicker::modeSave);

    // if we have a non-null filename use it, otherwise use default save message one
    if (aMsgFilename.IsEmpty())
    {
      nsString saveMsgStr;
      GetString(NS_LITERAL_STRING("defaultSaveMessageAsFileName"), saveMsgStr);
      filePicker->SetDefaultString(saveMsgStr);
    }
    else
      filePicker->SetDefaultString(aMsgFilename);

    // because we will be using GetFilterIndex()
    // we must call AppendFilters() one at a time,
    // in MESSENGER_SAVEAS_FILE_TYPE order
    nsString emlFilesStr;
    GetString(NS_LITERAL_STRING("EMLFiles"), emlFilesStr);
    filePicker->AppendFilter(emlFilesStr,
                             NS_LITERAL_STRING("*.eml"));
    filePicker->AppendFilters(nsIFilePicker::filterHTML);
    filePicker->AppendFilters(nsIFilePicker::filterText);
    filePicker->AppendFilters(nsIFilePicker::filterAll);

    // by default, set the filter type as "all"
    // because the default string is "message.eml"
    // and type is "all", we will use the filename extension, and will save as rfc/822 (.eml)
    // but if the user just changes the name (to .html or .txt), we will save as those types
    // if the type is "all".
    // http://bugzilla.mozilla.org/show_bug.cgi?id=96134#c23
    filePicker->SetFilterIndex(ANY_FILE_TYPE);

    PRInt16 dialogResult;

    nsCOMPtr <nsILocalFile> lastSaveDir;
    rv = GetLastSaveDirectory(getter_AddRefs(lastSaveDir));
    if (NS_SUCCEEDED(rv) && lastSaveDir)
      filePicker->SetDisplayDirectory(lastSaveDir);

    nsCOMPtr<nsILocalFile> localFile;
    nsAutoString fileName;
    rv = filePicker->Show(&dialogResult);
    if (NS_FAILED(rv) || dialogResult == nsIFilePicker::returnCancel)
      goto done;

    rv = filePicker->GetFile(getter_AddRefs(localFile));
    if (NS_FAILED(rv))
      goto done;

    if (dialogResult == nsIFilePicker::returnReplace) {
      // be extra safe and only delete when the file is really a file
      PRBool isFile;
      rv = localFile->IsFile(&isFile);
      if (NS_SUCCEEDED(rv) && isFile) {
        rv = localFile->Remove(PR_FALSE /* recursive delete */);
        NS_ENSURE_SUCCESS(rv, rv);
      }
    }

    rv = SetLastSaveDirectory(localFile);
    if (NS_FAILED(rv))
      goto done;

    rv = filePicker->GetFilterIndex(&saveAsFileType);
    if (NS_FAILED(rv))
      goto done;

    rv = localFile->GetLeafName(fileName);
    NS_ENSURE_SUCCESS(rv, rv);

    switch ( saveAsFileType )
    {
      // Add the right extenstion based on filter index and build a new localFile.
      case HTML_FILE_TYPE:
          if ( (fileName.RFind(HTML_FILE_EXTENSION, PR_TRUE, -1, sizeof(HTML_FILE_EXTENSION)-1) == kNotFound) &&
               (fileName.RFind(HTML_FILE_EXTENSION2, PR_TRUE, -1, sizeof(HTML_FILE_EXTENSION2)-1) == kNotFound) ) {
           fileName.AppendLiteral(HTML_FILE_EXTENSION2);
           localFile->SetLeafName(fileName);
        }
        break;
      case TEXT_FILE_TYPE:
        if (fileName.RFind(TEXT_FILE_EXTENSION, PR_TRUE, -1, sizeof(TEXT_FILE_EXTENSION)-1) == kNotFound) {
         fileName.AppendLiteral(TEXT_FILE_EXTENSION);
         localFile->SetLeafName(fileName);
        }
        break;
      case EML_FILE_TYPE:
        if (fileName.RFind(EML_FILE_EXTENSION, PR_TRUE, -1, sizeof(EML_FILE_EXTENSION)-1) == kNotFound) {
         fileName.AppendLiteral(EML_FILE_EXTENSION);
         localFile->SetLeafName(fileName);
        }
        break;
      case ANY_FILE_TYPE:
      default:
        // If no extension found then default it to .eml. Otherwise,
        // set the right file type based on the specified extension.
        PRBool noExtensionFound = PR_FALSE;
        if (fileName.RFind(".", 1) != kNotFound)
        {
          if ( (fileName.RFind(HTML_FILE_EXTENSION, PR_TRUE, -1, sizeof(HTML_FILE_EXTENSION)-1) != kNotFound) ||
               (fileName.RFind(HTML_FILE_EXTENSION2, PR_TRUE, -1, sizeof(HTML_FILE_EXTENSION2)-1) != kNotFound) )
            saveAsFileType = HTML_FILE_TYPE;
          else if (fileName.RFind(TEXT_FILE_EXTENSION, PR_TRUE, -1, sizeof(TEXT_FILE_EXTENSION)-1) != kNotFound)
            saveAsFileType = TEXT_FILE_TYPE;
          else if (fileName.RFind(EML_FILE_EXTENSION, PR_TRUE, -1, sizeof(EML_FILE_EXTENSION)-1) != kNotFound)
            saveAsFileType = EML_FILE_TYPE;
          else
            noExtensionFound = PR_TRUE;
        }
        else
          noExtensionFound = PR_TRUE;
        // Set default file type here.
        if (noExtensionFound)
        {
          saveAsFileType = EML_FILE_TYPE;
          fileName.AppendLiteral(EML_FILE_EXTENSION);
          localFile->SetLeafName(fileName);
        }
        break;
    }

    // XXX todo
    // document the ownership model of saveListener
    saveListener = new nsSaveMsgListener(localFile, this, nsnull);
    if (!saveListener) {
      rv = NS_ERROR_OUT_OF_MEMORY;
      goto done;
    }
    NS_ADDREF(saveListener);

    rv = saveListener->QueryInterface(NS_GET_IID(nsIUrlListener), getter_AddRefs(urlListener));
    if (NS_FAILED(rv))
      goto done;

    if (saveAsFileType == EML_FILE_TYPE)
    {
      rv = messageService->SaveMessageToDisk(PromiseFlatCString(aURI).get(), localFile, PR_FALSE,
        urlListener, nsnull,
        PR_TRUE, mMsgWindow);
    }
    else
    {
      nsCAutoString urlString(aURI);

      // we can't go RFC822 to TXT until bug #1775 is fixed
      // so until then, do the HTML to TXT conversion in
      // nsSaveMsgListener::OnStopRequest(), see ConvertBufToPlainText()
      //
      // Setup the URL for a "Save As..." Operation...
      // For now, if this is a save as TEXT operation, then do
      // a "printing" operation
      if (saveAsFileType == TEXT_FILE_TYPE)
      {
        saveListener->m_outputFormat = nsSaveMsgListener::ePlainText;
        saveListener->m_doCharsetConversion = PR_TRUE;
        urlString.AppendLiteral("?header=print");
      }
      else
      {
        saveListener->m_outputFormat = nsSaveMsgListener::eHTML;
        saveListener->m_doCharsetConversion = PR_FALSE;
        urlString.AppendLiteral("?header=saveas");
      }

      rv = CreateStartupUrl(urlString.get(), getter_AddRefs(url));
      NS_ASSERTION(NS_SUCCEEDED(rv), "CreateStartupUrl failed");
      if (NS_FAILED(rv))
        goto done;

      saveListener->m_channel = nsnull;
      rv = NS_NewInputStreamChannel(getter_AddRefs(saveListener->m_channel),
        url,
        nsnull);                // inputStream
      NS_ASSERTION(NS_SUCCEEDED(rv), "NS_NewInputStreamChannel failed");
      if (NS_FAILED(rv))
        goto done;

      nsCOMPtr<nsIStreamConverterService> streamConverterService = do_GetService("@mozilla.org/streamConverters;1");
      nsCOMPtr<nsISupports> channelSupport = do_QueryInterface(saveListener->m_channel);

      // we can't go RFC822 to TXT until bug #1775 is fixed
      // so until then, do the HTML to TXT conversion in
      // nsSaveMsgListener::OnStopRequest(), see ConvertBufToPlainText()
      rv = streamConverterService->AsyncConvertData(MESSAGE_RFC822,
        TEXT_HTML,
        saveListener,
        channelSupport,
        getter_AddRefs(convertedListener));
      NS_ASSERTION(NS_SUCCEEDED(rv), "AsyncConvertData failed");
      if (NS_FAILED(rv))
        goto done;

      rv = messageService->DisplayMessage(urlString.get(), convertedListener, mMsgWindow,
        nsnull, nsnull, nsnull);
    }
  }
  else
  {
    // ** save as Template
    nsCOMPtr <nsIFile> tmpTempFile;
    nsresult rv = GetSpecialDirectoryWithFileName(NS_OS_TEMP_DIR,
                                                  "nsmail.tmp",
                                                  getter_AddRefs(tmpTempFile));

    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr <nsILocalFile> tmpFile = do_QueryInterface(tmpTempFile, &rv);
    // For temp file, we should use restrictive 00600 instead of ATTACHMENT_PERMISSION 
    rv = tmpFile->CreateUnique(nsIFile::NORMAL_FILE_TYPE, 00600);
    if (NS_FAILED(rv)) goto done;

    // XXX todo
    // document the ownership model of saveListener
    saveListener = new nsSaveMsgListener(tmpFile, this, nsnull);
    if (!saveListener) {
      rv = NS_ERROR_OUT_OF_MEMORY;
      goto done;
    }
    NS_ADDREF(saveListener);

    if (aIdentity)
      rv = aIdentity->GetStationeryFolder(saveListener->m_templateUri);
    if (NS_FAILED(rv))
      goto done;

    PRBool needDummyHeader = StringBeginsWith(saveListener->m_templateUri, NS_LITERAL_CSTRING("mailbox://"));
    PRBool canonicalLineEnding = StringBeginsWith(saveListener->m_templateUri, NS_LITERAL_CSTRING("imap://"));

    rv = saveListener->QueryInterface(
      NS_GET_IID(nsIUrlListener),
      getter_AddRefs(urlListener));
    if (NS_FAILED(rv))
      goto done;

    rv = messageService->SaveMessageToDisk(PromiseFlatCString(aURI).get(), tmpFile,
      needDummyHeader,
      urlListener, nsnull,
      canonicalLineEnding, mMsgWindow);
  }

done:
  if (NS_FAILED(rv))
  {
    // XXX todo
    // document the ownership model of saveListener
    NS_IF_RELEASE(saveListener);
    Alert("saveMessageFailed");
  }
  return rv;
}

/**
 * Show a Save All dialog allowing the user to pick which folder to save
 * messages to.
 * @param [out] aSaveDir directory to save to. Will be null on cancel.
 */
nsresult
nsMessenger::GetSaveToDir(nsILocalFile **aSaveDir)
{
  nsresult rv;
  nsCOMPtr<nsIFilePicker> filePicker =
    do_CreateInstance("@mozilla.org/filepicker;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsString chooseFolderStr;
  GetString(NS_LITERAL_STRING("ChooseFolder"), chooseFolderStr);
  filePicker->Init(mWindow, chooseFolderStr, nsIFilePicker::modeGetFolder);

  nsCOMPtr<nsILocalFile> lastSaveDir;
  rv = GetLastSaveDirectory(getter_AddRefs(lastSaveDir));
  if (NS_SUCCEEDED(rv) && lastSaveDir)
    filePicker->SetDisplayDirectory(lastSaveDir);

  PRInt16 dialogResult;
  rv = filePicker->Show(&dialogResult);
  if (NS_FAILED(rv) || dialogResult == nsIFilePicker::returnCancel)
  {
    // We'll indicate this by setting the outparam to null.
    *aSaveDir = nsnull;
    return NS_OK;
  }

  nsCOMPtr<nsILocalFile> dir;
  rv = filePicker->GetFile(getter_AddRefs(dir));
  NS_ENSURE_SUCCESS(rv, rv);

  rv = SetLastSaveDirectory(dir);
  NS_ENSURE_SUCCESS(rv, rv);

  *aSaveDir = nsnull;
  dir.swap(*aSaveDir);
  return NS_OK;
}

NS_IMETHODIMP
nsMessenger::SaveMessages(PRUint32 aCount,
                          const PRUnichar **aFilenameArray,
                          const char **aMessageUriArray)
{
  NS_ENSURE_ARG_MIN(aCount, 1);
  NS_ENSURE_ARG_POINTER(aFilenameArray);
  NS_ENSURE_ARG_POINTER(aMessageUriArray);

  nsresult rv;

  nsCOMPtr<nsILocalFile> saveDir;
  rv = GetSaveToDir(getter_AddRefs(saveDir));
  NS_ENSURE_SUCCESS(rv, rv);
  if (!saveDir) // A null saveDir means that the user canceled the save.
    return NS_OK;

  for (PRUint32 i = 0; i < aCount; i++) {
    if (!aFilenameArray[i]) // just to be sure
      return NS_ERROR_FAILURE;

    nsCOMPtr<nsILocalFile> saveToFile = do_CreateInstance(NS_LOCAL_FILE_CONTRACTID, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    rv = saveToFile->InitWithFile(saveDir);
    NS_ENSURE_SUCCESS(rv, rv);

    rv = saveToFile->Append(nsDependentString(aFilenameArray[i]));
    NS_ENSURE_SUCCESS(rv, rv);

    rv = PromptIfFileExists(saveToFile);
    if (NS_FAILED(rv))
      continue;

    nsCOMPtr<nsIMsgMessageService> messageService;
    nsCOMPtr<nsIUrlListener> urlListener;

    rv = GetMessageServiceFromURI(nsDependentCString(aMessageUriArray[i]),
                                  getter_AddRefs(messageService));
    if (NS_FAILED(rv)) {
      Alert("saveMessageFailed");
      return rv;
    }

    nsSaveMsgListener *saveListener = new nsSaveMsgListener(saveToFile, this, nsnull);
    if (!saveListener) {
      NS_IF_RELEASE(saveListener);
      Alert("saveMessageFailed");
      return NS_ERROR_OUT_OF_MEMORY;
    }
    NS_ADDREF(saveListener);

    rv = saveListener->QueryInterface(NS_GET_IID(nsIUrlListener),
                                      getter_AddRefs(urlListener));
    if (NS_FAILED(rv)) {
      NS_IF_RELEASE(saveListener);
      Alert("saveMessageFailed");
      return rv;
    }

    // Ok, now save the message.
    rv = messageService->SaveMessageToDisk(aMessageUriArray[i],
                                           saveToFile, PR_FALSE,
                                           urlListener, nsnull,
                                           PR_TRUE, mMsgWindow);
  }
  return rv;
}

nsresult
nsMessenger::Alert(const char *stringName)
{
  nsresult rv = NS_OK;

  if (mDocShell)
  {
    nsCOMPtr<nsIPrompt> dialog(do_GetInterface(mDocShell));

    if (dialog)
    {
      nsString alertStr;
      GetString(NS_ConvertASCIItoUTF16(stringName), alertStr);
      rv = dialog->Alert(nsnull, alertStr.get());
    }
  }
  return rv;
}

nsresult
nsMessenger::DoCommand(nsIRDFCompositeDataSource* db, const nsACString& command,
                       nsISupportsArray *srcArray,
                       nsISupportsArray *argumentArray)
{
  nsresult rv;

  nsCOMPtr<nsIRDFService> rdfService(do_GetService(kRDFServiceCID, &rv));
  if(NS_FAILED(rv))
    return rv;

  nsCOMPtr<nsIRDFResource> commandResource;
  rv = rdfService->GetResource(command, getter_AddRefs(commandResource));
  if(NS_SUCCEEDED(rv))
    rv = db->DoCommand(srcArray, commandResource, argumentArray);

  return rv;
}

NS_IMETHODIMP
nsMessenger::MessageServiceFromURI(const nsACString& aUri, nsIMsgMessageService **aMsgService)
{
  NS_ENSURE_ARG_POINTER(aMsgService);
  return GetMessageServiceFromURI(aUri, aMsgService);
}

NS_IMETHODIMP
nsMessenger::MsgHdrFromURI(const nsACString& aUri, nsIMsgDBHdr **aMsgHdr)
{
  NS_ENSURE_ARG_POINTER(aMsgHdr);
  nsCOMPtr <nsIMsgMessageService> msgService;
  nsresult rv;

  if (mMsgWindow && (StringBeginsWith(aUri, NS_LITERAL_CSTRING("file:")) ||
                     nsDependentCString(aUri).Find("type=application/x-message-display") >= 0))
  {
    nsCOMPtr <nsIMsgHeaderSink> headerSink;
    mMsgWindow->GetMsgHeaderSink(getter_AddRefs(headerSink));
    if (headerSink)
    {
      rv = headerSink->GetDummyMsgHeader(aMsgHdr);
      // Is there a way to check if they're asking for the hdr currently
      // displayed in a stand-alone msg window from a .eml file?
      // (pretty likely if this is a file: uri)
      return rv;
    }
  }

  rv = GetMessageServiceFromURI(aUri, getter_AddRefs(msgService));
  NS_ENSURE_SUCCESS(rv, rv);
  return msgService->MessageURIToMsgHdr(PromiseFlatCString(aUri).get(), aMsgHdr);
}

NS_IMETHODIMP nsMessenger::GetUndoTransactionType(PRUint32 *txnType)
{
  NS_ENSURE_TRUE(txnType && mTxnMgr, NS_ERROR_NULL_POINTER);

  nsresult rv;
  *txnType = nsMessenger::eUnknown;
  nsCOMPtr<nsITransaction> txn;
  rv = mTxnMgr->PeekUndoStack(getter_AddRefs(txn));
  if (NS_SUCCEEDED(rv) && txn)
  {
    nsCOMPtr <nsIPropertyBag2> propertyBag = do_QueryInterface(txn, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    return propertyBag->GetPropertyAsUint32(NS_LITERAL_STRING("type"), txnType);
  }
  return rv;
}

NS_IMETHODIMP nsMessenger::CanUndo(PRBool *bValue)
{
  NS_ENSURE_TRUE(bValue && mTxnMgr, NS_ERROR_NULL_POINTER);

  nsresult rv;
  *bValue = PR_FALSE;
  PRInt32 count = 0;
  rv = mTxnMgr->GetNumberOfUndoItems(&count);
  if (NS_SUCCEEDED(rv) && count > 0)
    *bValue = PR_TRUE;
  return rv;
}

NS_IMETHODIMP nsMessenger::GetRedoTransactionType(PRUint32 *txnType)
{
  NS_ENSURE_TRUE(txnType && mTxnMgr, NS_ERROR_NULL_POINTER);

  nsresult rv;
  *txnType = nsMessenger::eUnknown;
  nsCOMPtr<nsITransaction> txn;
  rv = mTxnMgr->PeekRedoStack(getter_AddRefs(txn));
  if (NS_SUCCEEDED(rv) && txn)
  {
    nsCOMPtr <nsIPropertyBag2> propertyBag = do_QueryInterface(txn, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    return propertyBag->GetPropertyAsUint32(NS_LITERAL_STRING("type"), txnType);
  }
  return rv;
}

NS_IMETHODIMP nsMessenger::CanRedo(PRBool *bValue)
{
  NS_ENSURE_TRUE(bValue && mTxnMgr, NS_ERROR_NULL_POINTER);

  nsresult rv;
  *bValue = PR_FALSE;
  PRInt32 count = 0;
  rv = mTxnMgr->GetNumberOfRedoItems(&count);
  if (NS_SUCCEEDED(rv) && count > 0)
    *bValue = PR_TRUE;
  return rv;
}

NS_IMETHODIMP
nsMessenger::Undo(nsIMsgWindow *msgWindow)
{
  nsresult rv = NS_OK;
  if (mTxnMgr)
  {
    PRInt32 numTxn = 0;
    rv = mTxnMgr->GetNumberOfUndoItems(&numTxn);
    if (NS_SUCCEEDED(rv) && numTxn > 0)
    {
        nsCOMPtr<nsITransaction> txn;
        rv = mTxnMgr->PeekUndoStack(getter_AddRefs(txn));
        if (NS_SUCCEEDED(rv) && txn)
        {
            static_cast<nsMsgTxn*>(static_cast<nsITransaction*>(txn.get()))->SetMsgWindow(msgWindow);
        }
        mTxnMgr->UndoTransaction();
    }
  }
  return rv;
}

NS_IMETHODIMP
nsMessenger::Redo(nsIMsgWindow *msgWindow)
{
  nsresult rv = NS_OK;
  if (mTxnMgr)
  {
    PRInt32 numTxn = 0;
    rv = mTxnMgr->GetNumberOfRedoItems(&numTxn);
    if (NS_SUCCEEDED(rv) && numTxn > 0)
    {
        nsCOMPtr<nsITransaction> txn;
        rv = mTxnMgr->PeekRedoStack(getter_AddRefs(txn));
        if (NS_SUCCEEDED(rv) && txn)
        {
            static_cast<nsMsgTxn*>(static_cast<nsITransaction*>(txn.get()))->SetMsgWindow(msgWindow);
        }
        mTxnMgr->RedoTransaction();
    }
  }
  return rv;
}

NS_IMETHODIMP
nsMessenger::GetTransactionManager(nsITransactionManager* *aTxnMgr)
{
  NS_ENSURE_TRUE(mTxnMgr && aTxnMgr, NS_ERROR_NULL_POINTER);

  *aTxnMgr = mTxnMgr;
  NS_ADDREF(*aTxnMgr);

  return NS_OK;
}

NS_IMETHODIMP nsMessenger::SetDocumentCharset(const nsACString& aCharacterSet)
{
  // We want to redisplay the currently selected message (if any) but forcing the
  // redisplay to use characterSet
  if (!mLastDisplayURI.IsEmpty())
  {
    SetDisplayCharset(NS_LITERAL_CSTRING("UTF-8"));

    nsCOMPtr <nsIMsgMessageService> messageService;
    nsresult rv = GetMessageServiceFromURI(mLastDisplayURI, getter_AddRefs(messageService));

    if (NS_SUCCEEDED(rv) && messageService)
    {
      messageService->DisplayMessage(mLastDisplayURI.get(), mDocShell, mMsgWindow, nsnull,
                                     PromiseFlatCString(aCharacterSet).get(), nsnull);
    }
  }

  return NS_OK;
}

NS_IMETHODIMP
nsMessenger::GetLastDisplayedMessageUri(nsACString& aLastDisplayedMessageUri)
{
  aLastDisplayedMessageUri = mLastDisplayURI;
  return NS_OK;
}

nsSaveMsgListener::nsSaveMsgListener(nsIFile* aFile, nsMessenger *aMessenger, nsIUrlListener *aListener)
{
  m_file = do_QueryInterface(aFile);
  m_messenger = aMessenger;
  mListener = aListener;
  mUrlHasStopped = PR_FALSE;
  mRequestHasStopped = PR_FALSE;
  
    // rhp: for charset handling
  m_doCharsetConversion = PR_FALSE;
  m_saveAllAttachmentsState = nsnull;
  mProgress = 0;
  mMaxProgress = -1;
  mCanceled = PR_FALSE;
  m_outputFormat = eUnknown;
  mInitialized = PR_FALSE;
  m_dataBuffer = (char*) PR_CALLOC(FOUR_K+1);
}

nsSaveMsgListener::~nsSaveMsgListener()
{
}

//
// nsISupports
//
NS_IMPL_ISUPPORTS4(nsSaveMsgListener, nsIUrlListener, nsIMsgCopyServiceListener, nsIStreamListener, nsICancelable)


NS_IMETHODIMP
nsSaveMsgListener::Cancel(nsresult status)
{
  mCanceled = PR_TRUE;
  return NS_OK;
}

//
// nsIUrlListener
//
NS_IMETHODIMP
nsSaveMsgListener::OnStartRunningUrl(nsIURI* url)
{
  if (mListener)
    mListener->OnStartRunningUrl(url);
  return NS_OK;
}

NS_IMETHODIMP
nsSaveMsgListener::OnStopRunningUrl(nsIURI *url, nsresult exitCode)
{
  nsresult rv = exitCode;
  PRBool killSelf = PR_TRUE;
  mUrlHasStopped = PR_TRUE;

  if (m_outputStream)
  {
    if (mRequestHasStopped || !m_templateUri.IsEmpty())
    {
      m_outputStream->Close();
      m_outputStream = nsnull;
    }
    if (NS_FAILED(rv)) goto done;
    
    // ** save as template goes here
    if (!m_templateUri.IsEmpty()) 
    {
      nsCOMPtr<nsIRDFService> rdf(do_GetService(kRDFServiceCID, &rv));
      if (NS_FAILED(rv)) goto done;
      nsCOMPtr<nsIRDFResource> res;
      rv = rdf->GetResource(m_templateUri, getter_AddRefs(res));
      if (NS_FAILED(rv)) goto done;
      nsCOMPtr<nsIMsgFolder> templateFolder;
      templateFolder = do_QueryInterface(res, &rv);
      if (NS_FAILED(rv)) goto done;
      nsCOMPtr<nsIMsgCopyService> copyService = do_GetService(NS_MSGCOPYSERVICE_CONTRACTID);
      if (copyService)
      {
        nsCOMPtr <nsIFile> clone;
        m_file->Clone(getter_AddRefs(clone));
        rv = copyService->CopyFileMessage(clone, templateFolder, 
                                          MSG_FLAG_READ | MSG_FLAG_TEMPLATE, "", this, nsnull);
      }
      killSelf = PR_FALSE;
    }
  }

done:
  if (NS_FAILED(rv))
  {
    if (m_file)
      m_file->Remove(PR_FALSE);
    if (m_messenger)
        m_messenger->Alert("saveMessageFailed");
  }
  
  if (mRequestHasStopped && mListener)
    mListener->OnStopRunningUrl(url, exitCode);
  else
    mListenerUri = url;
  
  return rv;
}

NS_IMETHODIMP
nsSaveMsgListener::OnStartCopy(void)
{
  return NS_OK;
}

NS_IMETHODIMP
nsSaveMsgListener::OnProgress(PRUint32 aProgress, PRUint32 aProgressMax)
{
  return NS_OK;
}

NS_IMETHODIMP
nsSaveMsgListener::SetMessageKey(PRUint32 aKey)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsSaveMsgListener::GetMessageId(nsACString& aMessageId)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsSaveMsgListener::OnStopCopy(nsresult aStatus)
{
  if (m_file)
    m_file->Remove(PR_FALSE);
  Release(); // all done kill ourself
  return aStatus;
}

// initializes the progress window if we are going to show one
// and for OSX, sets creator flags on the output file
nsresult nsSaveMsgListener::InitializeDownload(nsIRequest * aRequest, PRUint32 aBytesDownloaded)
{
  nsresult rv = NS_OK;
  
  mInitialized = PR_TRUE;
  nsCOMPtr<nsIChannel> channel (do_QueryInterface(aRequest));
  
  if (!channel)
    return rv;
  
  // Get the max progress from the URL if we haven't already got it.
  if (mMaxProgress == -1)
  {
    nsCOMPtr<nsIURI> uri;
    channel->GetURI(getter_AddRefs(uri));
    nsCOMPtr<nsIMsgMailNewsUrl> mailnewsUrl(do_QueryInterface(uri));
    if (mailnewsUrl)
      mailnewsUrl->GetMaxProgress(&mMaxProgress);
  }
  
  if (!m_contentType.IsEmpty())
  {
    nsCOMPtr<nsIMIMEService> mimeService (do_GetService(NS_MIMESERVICE_CONTRACTID));
    nsCOMPtr<nsIMIMEInfo> mimeinfo;
    
    mimeService->GetFromTypeAndExtension(m_contentType, EmptyCString(), getter_AddRefs(mimeinfo));
    
    nsCOMPtr<nsILocalFile> outputFile = do_QueryInterface(m_file);
    
    // create a download progress window
    // We don't want to show the progress dialog if the download is really small.
    // but what is a small download? Well that's kind of arbitrary
    // so make an arbitrary decision based on the content length of the
    // attachment -- show it if less than half of the download has completed

    // When we don't allow warnings, also don't show progress, as this
    //  is an environment (typically filters) where we don't want
    //  interruption.
    PRBool allowProgress = PR_TRUE;
    if (m_saveAllAttachmentsState)
      allowProgress = !m_saveAllAttachmentsState->m_withoutWarning;
    if (allowProgress && mMaxProgress != -1 &&
        mMaxProgress > aBytesDownloaded * 2)
    {
      nsCOMPtr<nsITransfer> tr = do_CreateInstance(NS_TRANSFER_CONTRACTID, &rv);
      if (tr && outputFile)
      {
        PRTime timeDownloadStarted = PR_Now();
        
        nsCOMPtr<nsIURI> outputURI;
        NS_NewFileURI(getter_AddRefs(outputURI), outputFile);
        
        nsCOMPtr<nsIURI> url;
        channel->GetURI(getter_AddRefs(url));
        rv = tr->Init(url, outputURI, EmptyString(), mimeinfo,
                      timeDownloadStarted, nsnull, this);
        
          // now store the web progresslistener
        mTransfer = tr;
      }
    }
    
#ifdef XP_MACOSX
    /* if we are saving an appledouble or applesingle attachment, we need to use an Apple File Decoder */
    if (m_contentType.LowerCaseEqualsLiteral(APPLICATION_APPLEFILE) ||
        m_contentType.LowerCaseEqualsLiteral(MULTIPART_APPLEDOUBLE))
    {
      nsCOMPtr<nsIAppleFileDecoder> appleFileDecoder = do_CreateInstance(NS_IAPPLEFILEDECODER_CONTRACTID, &rv);
      if (NS_SUCCEEDED(rv) && appleFileDecoder)
      {
        rv = appleFileDecoder->Initialize(m_outputStream, outputFile);
        if (NS_SUCCEEDED(rv))
          m_outputStream = do_QueryInterface(appleFileDecoder, &rv);
      }
    }
#endif // XP_MACOSX
  }
  return rv;
}

NS_IMETHODIMP
nsSaveMsgListener::OnStartRequest(nsIRequest* request, nsISupports* aSupport)
{
  if (m_file)
    MsgNewBufferedFileOutputStream(getter_AddRefs(m_outputStream), m_file, -1, ATTACHMENT_PERMISSION);
  if (!m_outputStream)
  {
    mCanceled = PR_TRUE;
    if (m_messenger)
      m_messenger->Alert("saveAttachmentFailed");
  }
  return NS_OK;
}

NS_IMETHODIMP
nsSaveMsgListener::OnStopRequest(nsIRequest* request, nsISupports* aSupport,
                                 nsresult status)
{
  nsresult rv = NS_OK;
  mRequestHasStopped = PR_TRUE;
  
  // rhp: If we are doing the charset conversion magic, this is different
  // processing, otherwise, its just business as usual.
  //
  if ( (m_doCharsetConversion) && (m_outputStream) )
  {
    char        *conBuf = nsnull;
    PRUint32    conLength = 0;
    
    // If we need text/plain, then we need to convert the HTML and then convert
    // to the systems charset
    //
    if (m_outputFormat == ePlainText)
    {
      ConvertBufToPlainText(m_msgBuffer);
      rv = nsMsgI18NSaveAsCharset(TEXT_PLAIN, nsMsgI18NFileSystemCharset(),
                                  m_msgBuffer.get(), &conBuf);
      if ( NS_SUCCEEDED(rv) && (conBuf) )
        conLength = strlen(conBuf);
    }
    
    if ( (NS_SUCCEEDED(rv)) && (conBuf) )
    {
      PRUint32      writeCount;
      rv = m_outputStream->Write(conBuf, conLength, &writeCount);
      if (conLength != writeCount)
        rv = NS_ERROR_FAILURE;
    }
    
    PR_FREEIF(conBuf);
  }
 
  if (m_outputStream)
  {
    m_outputStream->Close();
    m_outputStream = nsnull;
  }
  
  if (m_saveAllAttachmentsState)
  {
    m_saveAllAttachmentsState->m_curIndex++;
    if (!mCanceled && m_saveAllAttachmentsState->m_curIndex < m_saveAllAttachmentsState->m_count)
    {
      nsSaveAllAttachmentsState *state = m_saveAllAttachmentsState;
      PRUint32 i = state->m_curIndex;
      nsCString unescapedName;
      nsCOMPtr<nsILocalFile> localFile = do_CreateInstance(NS_LOCAL_FILE_CONTRACTID, &rv);
      if (NS_FAILED(rv)) goto done;
      rv = localFile->InitWithNativePath(nsDependentCString(state->m_directoryName));
      
      if (NS_FAILED(rv)) goto done;
      
      rv = ConvertAndSanitizeFileName(state->m_displayNameArray[i], nsnull,
                                      getter_Copies(unescapedName));
      if (NS_FAILED(rv))
        goto done;
      
      localFile->AppendNative(unescapedName);
      rv = m_messenger->PromptIfFileExists(localFile);
      if (NS_FAILED(rv)) goto done;
      // When we are running with no warnings (typically filters and other automatic
      //  uses), then don't prompt for duplicates, but create a unique file
      //  instead.
      if (!m_saveAllAttachmentsState->m_withoutWarning)
      {
        rv = m_messenger->PromptIfFileExists(localFile);
        if (NS_FAILED(rv)) goto done;
      }
      else
      {
        rv = localFile->CreateUnique(nsIFile::NORMAL_FILE_TYPE, ATTACHMENT_PERMISSION);
        if (NS_FAILED(rv)) goto done;
      }
      rv = m_messenger->SaveAttachment(localFile,
                                       nsDependentCString(state->m_urlArray[i]),
                                       nsDependentCString(state->m_messageUriArray[i]),
                                       nsDependentCString(state->m_contentTypeArray[i]),
                                       (void *)state, nsnull);
      done:
        if (NS_FAILED(rv))
        {
          delete state;
          m_saveAllAttachmentsState = nsnull;
        }
    }
    else
    {
          // check if we're saving attachments prior to detaching them.
      if (m_saveAllAttachmentsState->m_detachingAttachments && !mCanceled)
      {
        nsSaveAllAttachmentsState *state = m_saveAllAttachmentsState;
        m_messenger->DetachAttachments(state->m_count,
                                       (const char **) state->m_contentTypeArray,
                                       (const char **) state->m_urlArray,
                                       (const char **) state->m_displayNameArray,
                                       (const char **) state->m_messageUriArray,
                                       &state->m_savedFiles,
                                       state->m_withoutWarning);
      }

      delete m_saveAllAttachmentsState;
      m_saveAllAttachmentsState = nsnull;
    }
  }

  if(mTransfer)
  {
    mTransfer->OnProgressChange64(nsnull, nsnull, mMaxProgress, mMaxProgress, mMaxProgress, mMaxProgress);
    mTransfer->OnStateChange(nsnull, nsnull, nsIWebProgressListener::STATE_STOP |
      nsIWebProgressListener::STATE_IS_NETWORK, NS_OK);
    mTransfer = nsnull; // break any circular dependencies between the progress dialog and use
  }
  
  if (mUrlHasStopped && mListener)
    mListener->OnStopRunningUrl(mListenerUri, rv);

  return NS_OK;
}

NS_IMETHODIMP
nsSaveMsgListener::OnDataAvailable(nsIRequest* request,
                                  nsISupports* aSupport,
                                  nsIInputStream* inStream,
                                  PRUint32 srcOffset,
                                  PRUint32 count)
{
  nsresult rv = NS_ERROR_FAILURE;
  // first, check to see if we've been canceled....
  if (mCanceled) // then go cancel our underlying channel too
    return request->Cancel(NS_BINDING_ABORTED);
  
  if (!mInitialized)
    InitializeDownload(request, count);

  if (m_dataBuffer && m_outputStream)
  {
    mProgress += count;
    PRUint32 available, readCount, maxReadCount = FOUR_K;
    PRUint32 writeCount;
    rv = inStream->Available(&available);
    while (NS_SUCCEEDED(rv) && available)
    {
      if (maxReadCount > available)
        maxReadCount = available;
      rv = inStream->Read(m_dataBuffer, maxReadCount, &readCount);

      // rhp:
      // Ok, now we do one of two things. If we are sending out HTML, then
      // just write it to the HTML stream as it comes along...but if this is
      // a save as TEXT operation, we need to buffer this up for conversion
      // when we are done. When the stream converter for HTML-TEXT gets in place,
      // this magic can go away.
      //
      if (NS_SUCCEEDED(rv))
      {
        if ( (m_doCharsetConversion) && (m_outputFormat == ePlainText) )
          m_msgBuffer.Append(NS_ConvertUTF8toUTF16(Substring(m_dataBuffer, m_dataBuffer + readCount)));
        else
          rv = m_outputStream->Write(m_dataBuffer, readCount, &writeCount);

        available -= readCount;
      }
    }

    if (NS_SUCCEEDED(rv) && mTransfer) // Send progress notification.
      mTransfer->OnProgressChange64(nsnull, request, mProgress, mMaxProgress, mProgress, mMaxProgress);
  }
  return rv;
}

#define MESSENGER_STRING_URL       "chrome://messenger/locale/messenger.properties"

nsresult
nsMessenger::InitStringBundle()
{
  nsresult res = NS_OK;
  if (!mStringBundle)
  {
    const char propertyURL[] = MESSENGER_STRING_URL;
    nsCOMPtr<nsIStringBundleService> sBundleService =
             do_GetService(NS_STRINGBUNDLE_CONTRACTID, &res);
    if (NS_SUCCEEDED(res) && (nsnull != sBundleService))
      res = sBundleService->CreateBundle(propertyURL,
                                               getter_AddRefs(mStringBundle));
  }
  return res;
}

void
nsMessenger::GetString(const nsString& aStringName, nsString& aValue)
{
  nsresult rv;
  aValue.Truncate();

  if (!mStringBundle)
    rv = InitStringBundle();

  if (mStringBundle)
    rv = mStringBundle->GetStringFromName(aStringName.get(), getter_Copies(aValue));

  if (NS_FAILED(rv) || aValue.IsEmpty())
    aValue = aStringName;
  return;
}

nsSaveAllAttachmentsState::nsSaveAllAttachmentsState(PRUint32 count,
                                                     const char **contentTypeArray,
                                                     const char **urlArray,
                                                     const char **nameArray,
                                                     const char **uriArray,
                                                     const char *dirName,
                                                     PRBool detachingAttachments)
    : m_withoutWarning(PR_FALSE)
{
    PRUint32 i;
    NS_ASSERTION(count && urlArray && nameArray && uriArray && dirName,
                 "fatal - invalid parameters\n");

    m_count = count;
    m_curIndex = 0;
    m_contentTypeArray = new char*[count];
    m_urlArray = new char*[count];
    m_displayNameArray = new char*[count];
    m_messageUriArray = new char*[count];
    for (i = 0; i < count; i++)
    {
        m_contentTypeArray[i] = strdup(contentTypeArray[i]);
        m_urlArray[i] = strdup(urlArray[i]);
        m_displayNameArray[i] = strdup(nameArray[i]);
        m_messageUriArray[i] = strdup(uriArray[i]);
    }
    m_directoryName = strdup(dirName);
    m_detachingAttachments = detachingAttachments;
}

nsSaveAllAttachmentsState::~nsSaveAllAttachmentsState()
{
    PRUint32 i;
    for (i = 0; i < m_count; i++)
    {
      NS_Free(m_contentTypeArray[i]);
      NS_Free(m_urlArray[i]);
      NS_Free(m_displayNameArray[i]);
      NS_Free(m_messageUriArray[i]);
    }
    delete[] m_contentTypeArray;
    delete[] m_urlArray;
    delete[] m_displayNameArray;
    delete[] m_messageUriArray;
    NS_Free(m_directoryName);
}

nsresult
nsMessenger::GetLastSaveDirectory(nsILocalFile **aLastSaveDir)
{
  nsresult rv;
  nsCOMPtr<nsIPrefBranch> prefBranch = do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv,rv);

  // this can fail, and it will, on the first time we call it, as there is no default for this pref.
  nsCOMPtr <nsILocalFile> localFile;
  rv = prefBranch->GetComplexValue(MESSENGER_SAVE_DIR_PREF_NAME, NS_GET_IID(nsILocalFile), getter_AddRefs(localFile));
  if (NS_SUCCEEDED(rv)) {
    NS_IF_ADDREF(*aLastSaveDir = localFile);
  }
  return rv;
}

nsresult
nsMessenger::SetLastSaveDirectory(nsILocalFile *aLocalFile)
{
  nsresult rv;
  nsCOMPtr<nsIPrefBranch> prefBranch = do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv,rv);

  nsCOMPtr <nsIFile> file = do_QueryInterface(aLocalFile, &rv);
  NS_ENSURE_SUCCESS(rv,rv);

  // if the file is a directory, just use it for the last dir chosen
  // otherwise, use the parent of the file as the last dir chosen.
  // IsDirectory() will return error on saving a file, as the
  // file doesn't exist yet.
  PRBool isDirectory;
  rv = file->IsDirectory(&isDirectory);
  if (NS_SUCCEEDED(rv) && isDirectory) {
    rv = prefBranch->SetComplexValue(MESSENGER_SAVE_DIR_PREF_NAME, NS_GET_IID(nsILocalFile), aLocalFile);
    NS_ENSURE_SUCCESS(rv,rv);
  }
  else {
    nsCOMPtr <nsIFile> parent;
    rv = file->GetParent(getter_AddRefs(parent));
    NS_ENSURE_SUCCESS(rv,rv);

    nsCOMPtr <nsILocalFile> parentLocalFile = do_QueryInterface(parent, &rv);
    NS_ENSURE_SUCCESS(rv,rv);

    rv = prefBranch->SetComplexValue(MESSENGER_SAVE_DIR_PREF_NAME, NS_GET_IID(nsILocalFile), parentLocalFile);
    NS_ENSURE_SUCCESS(rv,rv);
  }
  return NS_OK;
}

/* void getUrisAtNavigatePos (in long aPos, out ACString aFolderUri, out ACString aMsgUri); */
// aPos is relative to the current history cursor - 1 is forward, -1 is back.
NS_IMETHODIMP nsMessenger::GetMsgUriAtNavigatePos(PRInt32 aPos, nsACString& aMsgUri)
{
  PRInt32 desiredArrayIndex = (mCurHistoryPos + (aPos << 1));
  if (desiredArrayIndex >= 0 && desiredArrayIndex < mLoadedMsgHistory.Count())
  {
    mNavigatingToUri = mLoadedMsgHistory[desiredArrayIndex]->get();
    aMsgUri = mNavigatingToUri;
    return NS_OK;
  }
  return NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsMessenger::SetNavigatePos(PRInt32 aPos)
{
  if ((aPos << 1) < mLoadedMsgHistory.Count())
  {
    mCurHistoryPos = aPos << 1;
    return NS_OK;
  }
  else
    return NS_ERROR_INVALID_ARG;
}

NS_IMETHODIMP nsMessenger::GetNavigatePos(PRInt32 *aPos)
{
  NS_ENSURE_ARG_POINTER(aPos);
  *aPos = mCurHistoryPos >> 1;
  return NS_OK;
}

// aPos is relative to the current history cursor - 1 is forward, -1 is back.
NS_IMETHODIMP nsMessenger::GetFolderUriAtNavigatePos(PRInt32 aPos, nsACString& aFolderUri)
{
  PRInt32 desiredArrayIndex = (mCurHistoryPos + (aPos << 1));
  if (desiredArrayIndex >= 0 && desiredArrayIndex < mLoadedMsgHistory.Count())
  {
    mNavigatingToUri = mLoadedMsgHistory[desiredArrayIndex + 1]->get();
    aFolderUri = mNavigatingToUri;
    return NS_OK;
  }
  return NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsMessenger::GetNavigateHistory(PRUint32 *aCurPos, PRUint32 *aCount, char *** aHistoryUris)
{
  NS_ENSURE_ARG_POINTER(aCount);
  NS_ENSURE_ARG_POINTER(aCurPos);

  *aCurPos = mCurHistoryPos >> 1;
  *aCount = mLoadedMsgHistory.Count();
  // for just enabling commands, we don't need the history uris.
  if (!aHistoryUris)
    return NS_OK;

  char **outArray, **next;
  next = outArray = (char **)nsMemory::Alloc(*aCount * sizeof(char *));
  if (!outArray) return NS_ERROR_OUT_OF_MEMORY;
  for (PRUint32 i = 0; i < *aCount; i++)
  {
    *next = ToNewCString(*(mLoadedMsgHistory[i]));
    if (!*next)
      return NS_ERROR_OUT_OF_MEMORY;
    next++;
  }
  *aHistoryUris = outArray;
  return NS_OK;
}

NS_IMETHODIMP
nsMessenger::FormatFileSize(PRUint64 aSize, PRBool aUseKB, nsAString& aFormattedSize)
{
  return ::FormatFileSize(aSize, aUseKB, aFormattedSize);
}

/* void OnItemAdded (in nsIRDFResource parentItem, in nsISupports item); */
NS_IMETHODIMP nsMessenger::OnItemAdded(nsIRDFResource *parentItem, nsISupports *item)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemRemoved (in nsIRDFResource parentItem, in nsISupports item); */
NS_IMETHODIMP nsMessenger::OnItemRemoved(nsIRDFResource *parentItem, nsISupports *item)
{
  // check if this item is a message header that's in our history list. If so,
  // remove it from the history list.
  nsCOMPtr <nsIMsgDBHdr> msgHdr = do_QueryInterface(item);
  if (msgHdr)
  {
    nsCOMPtr <nsIMsgFolder> folder;
    msgHdr->GetFolder(getter_AddRefs(folder));
    if (folder)
    {
      nsCString msgUri;
      nsMsgKey msgKey;
      msgHdr->GetMessageKey(&msgKey);
      folder->GenerateMessageURI(msgKey, msgUri);
      // need to remove the correspnding folder entry, and
      // adjust the current history pos.
      PRInt32 uriPos = mLoadedMsgHistory.IndexOf(msgUri);
      if (uriPos != kNotFound)
      {
        mLoadedMsgHistory.RemoveCStringAt(uriPos);
        mLoadedMsgHistory.RemoveCStringAt(uriPos); // and the folder uri entry
        if ((PRInt32) mCurHistoryPos >= uriPos)
          mCurHistoryPos -= 2;
      }
    }
  }
  return NS_OK;
}

/* void OnItemPropertyChanged (in nsIRDFResource item, in nsIAtom property, in string oldValue, in string newValue); */
NS_IMETHODIMP nsMessenger::OnItemPropertyChanged(nsIRDFResource *item, nsIAtom *property, const char *oldValue, const char *newValue)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemIntPropertyChanged (in nsIRDFResource item, in nsIAtom property, in long oldValue, in long newValue); */
NS_IMETHODIMP nsMessenger::OnItemIntPropertyChanged(nsIRDFResource *item, nsIAtom *property, PRInt32 oldValue, PRInt32 newValue)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemBoolPropertyChanged (in nsIRDFResource item, in nsIAtom property, in boolean oldValue, in boolean newValue); */
NS_IMETHODIMP nsMessenger::OnItemBoolPropertyChanged(nsIRDFResource *item, nsIAtom *property, PRBool oldValue, PRBool newValue)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemUnicharPropertyChanged (in nsIRDFResource item, in nsIAtom property, in wstring oldValue, in wstring newValue); */
NS_IMETHODIMP nsMessenger::OnItemUnicharPropertyChanged(nsIRDFResource *item, nsIAtom *property, const PRUnichar *oldValue, const PRUnichar *newValue)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemPropertyFlagChanged (in nsIMsgDBHdr item, in nsIAtom property, in unsigned long oldFlag, in unsigned long newFlag); */
NS_IMETHODIMP nsMessenger::OnItemPropertyFlagChanged(nsIMsgDBHdr *item, nsIAtom *property, PRUint32 oldFlag, PRUint32 newFlag)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

/* void OnItemEvent (in nsIMsgFolder item, in nsIAtom event); */
NS_IMETHODIMP nsMessenger::OnItemEvent(nsIMsgFolder *item, nsIAtom *event)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

///////////////////////////////////////////////////////////////////////////////
// Detach/Delete Attachments
///////////////////////////////////////////////////////////////////////////////

static char * GetAttachmentPartId(const char * aAttachmentUrl)
{
  static const char partIdPrefix[] = "part=";
  char * partId = PL_strstr(aAttachmentUrl, partIdPrefix);
  return partId ? (partId + sizeof(partIdPrefix) - 1) : nsnull;
}

static int CompareAttachmentPartId(const char * aAttachUrlLeft, const char * aAttachUrlRight)
{
  // part ids are numbers separated by periods, like "1.2.3.4".
  // we sort by doing a numerical comparison on each item in turn. e.g. "1.4" < "1.25"
  // shorter entries come before longer entries. e.g. "1.4" < "1.4.1.2"
  // return values:
  //  -2  left is a parent of right
  //  -1  left is less than right
  //   0  left == right
  //   1  right is greater than left
  //   2  right is a parent of left

  char * partIdLeft  = GetAttachmentPartId(aAttachUrlLeft);
  char * partIdRight = GetAttachmentPartId(aAttachUrlRight);

  // for detached attachments the URL does not contain any "part=xx"
  if(!partIdLeft)
    partIdLeft = "0";

  if(!partIdRight)
    partIdRight = "0";

  long idLeft, idRight;
  do
  {
    NS_ABORT_IF_FALSE(partIdLeft && NS_IS_DIGIT(*partIdLeft), "Invalid character in part id string");
    NS_ABORT_IF_FALSE(partIdRight && NS_IS_DIGIT(*partIdRight), "Invalid character in part id string");

    // if the part numbers are different then the numerically smaller one is first
    idLeft  = strtol(partIdLeft, &partIdLeft, 10);
    idRight = strtol(partIdRight, &partIdRight, 10);
    if (idLeft != idRight)
      return idLeft < idRight ? -1 : 1;

    // if one part id is complete but the other isn't, then the shortest one
    // is first (parents before children)
    if (*partIdLeft != *partIdRight)
      return *partIdRight ? -2 : 2;

    // if both part ids are complete (*partIdLeft == *partIdRight now) then
    // they are equal
    if (!*partIdLeft)
      return 0;

    NS_ABORT_IF_FALSE(*partIdLeft == '.', "Invalid character in part id string");
    NS_ABORT_IF_FALSE(*partIdRight == '.', "Invalid character in part id string");

    ++partIdLeft;
    ++partIdRight;
  }
  while (PR_TRUE);
  return 0;
}

// ------------------------------------

// struct on purpose -> show that we don't ever want a vtable
struct msgAttachment
{
  msgAttachment()
    : mContentType(nsnull),
      mUrl(nsnull),
      mDisplayName(nsnull),
      mMessageUri(nsnull)
  {
  }

  ~msgAttachment()
  {
    Clear();
  }

  void Clear()
  {
    NS_Free(mContentType);
    NS_Free(mUrl);
    NS_Free(mDisplayName);
    NS_Free(mMessageUri);
  }

  PRBool Init(const char * aContentType, const char * aUrl,
              const char * aDisplayName, const char * aMessageUri)
  {
    Clear();
    mContentType = strdup(aContentType);
    mUrl = strdup(aUrl);
    mDisplayName = strdup(aDisplayName);
    mMessageUri = strdup(aMessageUri);
    return (mContentType && mUrl && mDisplayName && mMessageUri);
  }

  // take the pointers from aSource
  void Adopt(msgAttachment & aSource)
  {
    Clear();

    mContentType = aSource.mContentType;
    mUrl = aSource.mUrl;
    mDisplayName = aSource.mDisplayName;
    mMessageUri = aSource.mMessageUri;

    aSource.mContentType = nsnull;
    aSource.mUrl = nsnull;
    aSource.mDisplayName = nsnull;
    aSource.mMessageUri = nsnull;
  }

  char* mContentType;
  char* mUrl;
  char* mDisplayName;
  char* mMessageUri;

private:
  // disable by not implementing
  msgAttachment(const msgAttachment & rhs);
  msgAttachment & operator=(const msgAttachment & rhs);
};

// ------------------------------------

class nsAttachmentState
{
public:
  nsAttachmentState();
  ~nsAttachmentState();
  nsresult Init(PRUint32 aCount,
                const char **aContentTypeArray,
                const char **aUrlArray,
                const char **aDisplayNameArray,
                const char **aMessageUriArray);
  nsresult PrepareForAttachmentDelete();

private:
  static int SortAttachmentsByPartId(const void * aLeft, const void * aRight, void *);

public:
  PRUint32        mCount;
  PRUint32        mCurIndex;
  msgAttachment*  mAttachmentArray;
};

nsAttachmentState::nsAttachmentState()
  : mCount(0),
    mCurIndex(0),
    mAttachmentArray(nsnull)
{
}

nsAttachmentState::~nsAttachmentState()
{
  delete[] mAttachmentArray;
}

nsresult
nsAttachmentState::Init(PRUint32 aCount, const char ** aContentTypeArray,
                        const char ** aUrlArray, const char ** aDisplayNameArray,
                        const char ** aMessageUriArray)
{
  NS_ABORT_IF_FALSE(aCount > 0, "count is invalid");

  mCount = aCount;
  mCurIndex = 0;
  delete[] mAttachmentArray;

  mAttachmentArray = new msgAttachment[aCount];
  if (!mAttachmentArray)
    return NS_ERROR_OUT_OF_MEMORY;

  for(PRUint32 u = 0; u < aCount; ++u)
  {
    if (!mAttachmentArray[u].Init(aContentTypeArray[u], aUrlArray[u],
      aDisplayNameArray[u], aMessageUriArray[u]))
      return NS_ERROR_OUT_OF_MEMORY;
  }

  return NS_OK;
}

nsresult
nsAttachmentState::PrepareForAttachmentDelete()
{
  // this must be called before any processing
  if (mCurIndex != 0)
    return NS_ERROR_FAILURE;

  // this prepares the attachment list for use in deletion. In order to prepare, we
  // sort the attachments in numerical ascending order on their part id, remove all
  // duplicates and remove any subparts which will be removed automatically by the
  // removal of the parent.
  //
  // e.g. the attachment list processing (showing only part ids)
  // before: 1.11, 1.3, 1.2, 1.2.1.3, 1.4.1.2
  // sorted: 1.2, 1.2.1.3, 1.3, 1.4.1.2, 1.11
  // after:  1.2, 1.3, 1.4.1.2, 1.11

  // sort
  NS_QuickSort(mAttachmentArray, mCount, sizeof(msgAttachment), SortAttachmentsByPartId, nsnull);

  // remove duplicates and sub-items
  int nCompare;
  for(PRUint32 u = 1; u < mCount;)
  {
    nCompare = ::CompareAttachmentPartId(mAttachmentArray[u-1].mUrl, mAttachmentArray[u].mUrl);
    if (nCompare == 0 || nCompare == -2) // [u-1] is the same as or a parent of [u]
    {
      // shuffle the array down (and thus keeping the sorted order)
      // this will get rid of the current unnecessary element
      for (PRUint32 i = u + 1; i < mCount; ++i)
      {
        mAttachmentArray[i-1].Adopt(mAttachmentArray[i]);
      }
      --mCount;
    }
    else
    {
      ++u;
    }
  }

  return NS_OK;
}

int
nsAttachmentState::SortAttachmentsByPartId(const void * aLeft, const void * aRight, void *)
{
  msgAttachment & attachLeft  = *((msgAttachment*) aLeft);
  msgAttachment & attachRight = *((msgAttachment*) aRight);
  return ::CompareAttachmentPartId(attachLeft.mUrl, attachRight.mUrl);
}

// ------------------------------------

class nsDelAttachListener : public nsIStreamListener,
                            public nsIUrlListener,
                            public nsIMsgCopyServiceListener
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSISTREAMLISTENER
  NS_DECL_NSIREQUESTOBSERVER
  NS_DECL_NSIURLLISTENER
  NS_DECL_NSIMSGCOPYSERVICELISTENER

public:
  nsDelAttachListener();
  virtual ~nsDelAttachListener();
  nsresult StartProcessing(nsMessenger * aMessenger, nsIMsgWindow * aMsgWindow,
    nsAttachmentState * aAttach, PRBool aSaveFirst);
  nsresult DeleteOriginalMessage();
  void SelectNewMessage();
  nsresult FixHeadersAndCopyDataToOutputStream();

public:
  nsAttachmentState * mAttach;                      // list of attachments to process
  PRBool mSaveFirst;                                // detach (PR_TRUE) or delete (PR_FALSE)
  nsCOMPtr<nsIFile> mMsgFile;                       // temporary file (processed mail)
  nsCOMPtr<nsIOutputStream> mMsgFileStream;         // temporary file (processed mail)
  nsCOMPtr<nsIMsgMessageService> mMessageService;   // original message service
  nsCOMPtr<nsIMsgDBHdr> mOriginalMessage;           // original message header
  nsCOMPtr<nsIMsgFolder> mMessageFolder;            // original message folder
  nsCOMPtr<nsIMessenger> mMessenger;                // our messenger instance
  nsCOMPtr<nsIMsgWindow> mMsgWindow;                // our UI window
  PRUint32 mNewMessageKey;                          // new message key
  PRUint32 mOrigMsgFlags;
  nsCString mMessage;

   enum {
      eStarting,
      eCopyingNewMsg,
      eUpdatingFolder, // for IMAP
      eDeletingOldMessage,
      eSelectingNewMessage
    } m_state;
   // temp
  PRBool mWrittenExtra;
  PRBool mDetaching;
  nsCStringArray mDetachedFileUris;
};

//
// nsISupports
//
NS_IMPL_ISUPPORTS3(nsDelAttachListener,nsIStreamListener,nsIUrlListener,nsIMsgCopyServiceListener)

//
// nsIRequestObserver
//
NS_IMETHODIMP
nsDelAttachListener::OnStartRequest(nsIRequest * aRequest, nsISupports * aContext)
{
  // called when we start processing the StreamMessage request.
  // This is called after OnStartRunningUrl().
  return NS_OK;
}

nsresult
nsDelAttachListener::FixHeadersAndCopyDataToOutputStream()
{
  if (!mMsgFileStream)
    return NS_ERROR_NULL_POINTER;

  // we have both headers and body in mMessage
  // find the header/body delimiter of a blank line.  A blank line
  // looks like <CR><CR>, <LF><LF>, or <CRLF><CRLF>
  PRInt32 firstCRCR = mMessage.Find("\r\r");
  PRInt32 firstLFLF = mMessage.Find("\n\n");
  PRInt32 firstCRLFCRLF = mMessage.Find("\r\n\r\n");

  PRInt32 delimiterPos = kNotFound;
  PRInt32 delimiterLength = 2;

  // find the first blankline
  if (firstCRCR != kNotFound)
    delimiterPos = firstCRCR;
  if (firstLFLF != kNotFound && (delimiterPos == kNotFound || firstLFLF < delimiterPos))
    delimiterPos = firstLFLF;
  if (firstCRLFCRLF != kNotFound && (delimiterPos == kNotFound || firstCRLFCRLF < delimiterPos))
  {
    delimiterPos = firstCRLFCRLF;
    delimiterLength = 4;
  }
 
  nsCString headers;
  nsCString body;

  if (delimiterPos == kNotFound)
  {
    // is this even possible?  Assume that we have all headers and no body.
    headers = mMessage;
  }
  else
  {
    headers = Substring(mMessage, 0, delimiterPos);
    body = Substring(mMessage, delimiterPos + delimiterLength, mMessage.Length() - delimiterPos - delimiterLength);
  }

  // for other headers
  nsTArray<nsCString> allHeaders;
  
  // FIX ME scott broke this!
  ParseString(headers, '\n', allHeaders);
  nsCString fixedMessage;
  PRBool eatingHeader = PR_FALSE;
  for (PRInt32 i = 0; i < allHeaders.Length(); i++)
  {
    // Postbox Bug 6046 - since ParseString splits on \n remove the CR
    // from the header (if there is one) before we append our CRLF after each header.
    allHeaders[i].StripChars("\r");
    const char *currentHeader = allHeaders[i].get();

    if (!PL_strncasecmp(currentHeader, "Date:", strlen("Date:")))
    {
      eatingHeader = PR_TRUE;
    }
    else if (currentHeader[0] != ' ' || currentHeader[0] != '\t')
    {
      eatingHeader = PR_FALSE;
    }

    if (!eatingHeader)
    {
      fixedMessage += currentHeader;
      fixedMessage += CRLF;
    }
  }

  PRTime date;
  nsresult rv = mOriginalMessage->GetDate(&date);
  NS_ENSURE_SUCCESS(rv, rv);

  // twiddle the date forward by 1 second 
  // this way, new version of the message differs from the old version by a second.  
  // gmail, and possibly other imap servers, will get confused otherwise.
  // see bug #650, bug #817 and bug #907 for details
  date += PR_USEC_PER_SEC;

  // shamelessly copied from mime_generate_headers() in nsMsgCompUtils.cpp
  // code clean up one day, see bug #1212
  PRExplodedTime now;
  PR_ExplodeTime(date, PR_LocalTimeParameters, &now);

  char datetime[512];
  /* Use PR_FormatTimeUSEnglish() to format the date in US English format,
     then figure out what our local GMT offset is, and append it (since
     PR_FormatTimeUSEnglish() can't do that.) Generate four digit years as
     per RFC 1123 (superceding RFC 822.)
   */
  PR_FormatTimeUSEnglish(datetime, sizeof(datetime),
               "Date: %a, %d %b %Y %H:%M:%S ",
               &now);

  int gmtoffset = (now.tm_params.tp_gmt_offset + now.tm_params.tp_dst_offset) / 60;

  char gmtoffsetString[512];
  PR_snprintf(gmtoffsetString, sizeof(gmtoffsetString),
        "%c%02d%02d" CRLF,
        (gmtoffset >= 0 ? '+' : '-'),
        ((gmtoffset >= 0 ? gmtoffset : -gmtoffset) / 60),
        ((gmtoffset >= 0 ? gmtoffset : -gmtoffset) % 60));

  // create a stream with the headers, the new Date: header, CRLF, and the body
  fixedMessage += datetime;
  fixedMessage += gmtoffsetString;
  fixedMessage += CRLF;
  fixedMessage += body;

  nsCOMPtr<nsIStringInputStream> inputStream =
        do_CreateInstance("@mozilla.org/io/string-input-stream;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = inputStream->SetData(fixedMessage.get(), fixedMessage.Length());
  NS_ENSURE_SUCCESS(rv, rv);

  rv = mMessageFolder->CopyDataToOutputStreamForAppend(inputStream, fixedMessage.Length(), mMsgFileStream);
  NS_ENSURE_SUCCESS(rv, rv);
  return NS_OK;
}

NS_IMETHODIMP
nsDelAttachListener::OnStopRequest(nsIRequest * aRequest, nsISupports * aContext, nsresult aStatusCode)
{
  // called when we have completed processing the StreamMessage request.
  // This is called after OnStopRequest(). This means that we have now
  // received all data of the message and we have completed processing.
  // We now start to copy the processed message from the temporary file
  // back into the message store, replacing the original message.
  nsresult rv = FixHeadersAndCopyDataToOutputStream();
  NS_ENSURE_SUCCESS(rv,rv);

  mMessageFolder->CopyDataDone();
  if (NS_FAILED(aStatusCode))
    return aStatusCode;

  // called when we complete processing of the StreamMessage request.
  // This is called before OnStopRunningUrl().

  // copy the file back into the folder. Note: if we set msgToReplace then
  // CopyFileMessage() fails, do the delete ourselves
  nsCOMPtr<nsIMsgCopyServiceListener> listenerCopyService;
  rv = this->QueryInterface( NS_GET_IID(nsIMsgCopyServiceListener), getter_AddRefs(listenerCopyService) );
  NS_ENSURE_SUCCESS(rv,rv);

  mMsgFileStream->Close();
  mMsgFileStream = nsnull;
  mNewMessageKey = PR_UINT32_MAX;
  nsCOMPtr<nsIMsgCopyService> copyService = do_GetService(NS_MSGCOPYSERVICE_CONTRACTID);
  m_state = eCopyingNewMsg;
  // clone file because nsIFile on Windows caches the wrong file size.
  nsCOMPtr <nsIFile> clone;
  mMsgFile->Clone(getter_AddRefs(clone));
  if (copyService)
  {
    nsCString messageKeywords;
    mOriginalMessage->GetStringProperty("keywords", getter_Copies(messageKeywords));
    rv = copyService->CopyFileMessage(clone, mMessageFolder, mOrigMsgFlags,
                                      messageKeywords.get(), listenerCopyService, mMsgWindow);
  }
  return rv;
}

//
// nsIStreamListener
//

NS_IMETHODIMP
nsDelAttachListener::OnDataAvailable(nsIRequest * aRequest, nsISupports * aSupport,
                                     nsIInputStream * aInStream, PRUint32 aSrcOffset,
                                     PRUint32 aCount)
{
  // shameless copied from nsMsgAlterHelper::OnDataAvailable()
  // code clean up one day, see bug #1212

  // looking at nsPipe3.cpp, we have:
  // in nsPipe::Init() we initialize the mBuffer (a nsSegmentedBuffer) to
  // segmentSize * segmentCount (by default, that's 4096 * 16, or 65536)
  // so set the readBuf size to that amount (+1, for the required null byte)
  char readBuf[65537];

  PRUint32 available, readCount;
  PRUint32 maxReadCount = sizeof(readBuf) - 1;

  NS_ENSURE_ARG_POINTER(aInStream);

  nsresult rv = aInStream->Available(&available);
  while (NS_SUCCEEDED(rv) && available > 0)
  {
    memset(readBuf, 0, sizeof(readBuf));

    if (maxReadCount > available)
      maxReadCount = available;

    rv = aInStream->Read(readBuf, maxReadCount, &readCount);
    available -= readCount;
    mMessage.Append(readBuf);
  }
  return NS_OK;
}

//
// nsIUrlListener
//

NS_IMETHODIMP
nsDelAttachListener::OnStartRunningUrl(nsIURI * aUrl)
{
  // called when we start processing the StreamMessage request. This is
  // called before OnStartRequest().
  return NS_OK;
}

nsresult nsDelAttachListener::DeleteOriginalMessage()
{
  nsCOMPtr<nsISupportsArray> messageArray;
  nsresult rv = NS_NewISupportsArray(getter_AddRefs(messageArray));
  NS_ENSURE_SUCCESS(rv,rv);
  rv = messageArray->AppendElement(mOriginalMessage);
  NS_ENSURE_SUCCESS(rv,rv);
  nsCOMPtr<nsIMsgCopyServiceListener> listenerCopyService;

  QueryInterface( NS_GET_IID(nsIMsgCopyServiceListener), getter_AddRefs(listenerCopyService) );

  mOriginalMessage = nsnull;
  m_state = eDeletingOldMessage;
  return mMessageFolder->DeleteMessages(
    messageArray,         // messages
    mMsgWindow,           // msgWindow
    PR_TRUE,              // deleteStorage
    PR_FALSE,              // isMove
    listenerCopyService,  // listener
    PR_FALSE);            // allowUndo
}

void nsDelAttachListener::SelectNewMessage()
{
  nsCString displayUri;
  // all attachments refer to the same message
  const char * messageUri = mAttach->mAttachmentArray[0].mMessageUri;
  mMessenger->GetLastDisplayedMessageUri(displayUri);
  if (displayUri.Equals(messageUri))
  {
    mMessageFolder->GenerateMessageURI(mNewMessageKey, displayUri);
    if (!displayUri.IsEmpty() && mMsgWindow)
    {
      nsCOMPtr<nsIMsgWindowCommands> windowCommands;
      mMsgWindow->GetWindowCommands(getter_AddRefs(windowCommands));
      if (windowCommands)
        windowCommands->SelectMessage(displayUri);
    }
  }
  mNewMessageKey = PR_UINT32_MAX;
}

NS_IMETHODIMP
nsDelAttachListener::OnStopRunningUrl(nsIURI * aUrl, nsresult aExitCode)
{
  nsresult rv = NS_OK;
  const char * messageUri = mAttach->mAttachmentArray[0].mMessageUri;
  if (mOriginalMessage && !strncmp(messageUri, "imap-message:", 13))
  {
    if (m_state == eUpdatingFolder)
      rv = DeleteOriginalMessage();
  }
  // check if we've deleted the original message, and we know the new msg id.
  else if (m_state == eDeletingOldMessage && mMsgWindow)
    SelectNewMessage();

  return rv;
}

//
// nsIMsgCopyServiceListener
//

NS_IMETHODIMP
nsDelAttachListener::OnStartCopy(void)
{
  // never called?
  return NS_OK;
}

NS_IMETHODIMP
nsDelAttachListener::OnProgress(PRUint32 aProgress, PRUint32 aProgressMax)
{
  // never called?
  return NS_OK;
}

NS_IMETHODIMP
nsDelAttachListener::SetMessageKey(PRUint32 aKey)
{
  // called during the copy of the modified message back into the message
  // store to notify us of the message key of the newly created message.
  mNewMessageKey = aKey;
  return NS_OK;
}

NS_IMETHODIMP
nsDelAttachListener::GetMessageId(nsACString& aMessageId)
{
  // never called?
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDelAttachListener::OnStopCopy(nsresult aStatus)
{
  // only if the currently selected message is the one that we are about to delete then we
  // change the selection to the new message that we just added. Failures in this code are not fatal.
  // Note that can only do this if we have the new message key, which we don't always get from IMAP.
  // delete the original message
  if (NS_FAILED(aStatus))
    return aStatus;

  // check if we've deleted the original message, and we know the new msg id.
  if (m_state == eDeletingOldMessage && mMsgWindow)
    SelectNewMessage();
  // do this for non-imap messages - for imap, we'll do the delete in
  // OnStopRunningUrl. For local messages, we won't get an OnStopRunningUrl
  // notification. And for imap, it's too late to delete the message here,
  // because we'll be updating the folder naturally as a result of
  // running an append url. If we delete the header here, that folder
  // update will think we need to download the header...If we do it
  // in OnStopRunningUrl, we'll issue the delete before we do the
  // update....all nasty stuff.
  const char * messageUri = mAttach->mAttachmentArray[0].mMessageUri;
  if (mOriginalMessage && strncmp(messageUri, "imap-message:", 13))
    return DeleteOriginalMessage();
  else
    m_state = eUpdatingFolder;
  return NS_OK;
}

//
// local methods
//

nsDelAttachListener::nsDelAttachListener()
{
  mAttach = nsnull;
  mSaveFirst = PR_FALSE;
  mWrittenExtra = PR_FALSE;
  mNewMessageKey = PR_UINT32_MAX;
  m_state = eStarting;
}

nsDelAttachListener::~nsDelAttachListener()
{
  if (mAttach)
  {
    delete mAttach;
  }
  if (mMsgFileStream)
  {
    mMsgFileStream->Close();
    mMsgFileStream = nsnull;
  }
  if (mMsgFile)
  {
    mMsgFile->Remove(PR_FALSE);
  }
}

nsresult
nsDelAttachListener::StartProcessing(nsMessenger * aMessenger, nsIMsgWindow * aMsgWindow,
                                     nsAttachmentState * aAttach, PRBool detaching)
{
  aMessenger->QueryInterface(NS_GET_IID(nsIMessenger), getter_AddRefs(mMessenger));
  mMsgWindow = aMsgWindow;
  mAttach    = aAttach;
  mDetaching = detaching;

  nsresult rv;

  // all attachments refer to the same message
  const char * messageUri = mAttach->mAttachmentArray[0].mMessageUri;

  // get the message service, original message and folder for this message
  rv = GetMessageServiceFromURI(nsDependentCString(messageUri), getter_AddRefs(mMessageService));
  NS_ENSURE_SUCCESS(rv,rv);
  rv = mMessageService->MessageURIToMsgHdr(messageUri, getter_AddRefs(mOriginalMessage));
  NS_ENSURE_SUCCESS(rv,rv);

  // see bug #4478 for details
  NS_ENSURE_TRUE(mOriginalMessage, NS_ERROR_UNEXPECTED);

  rv = mOriginalMessage->GetFolder(getter_AddRefs(mMessageFolder));
  NS_ENSURE_SUCCESS(rv,rv);
  mOriginalMessage->GetFlags(&mOrigMsgFlags);

  // ensure that we can store and delete messages in this folder, if we
  // can't then we can't do attachment deleting
  PRBool canDelete = PR_FALSE;
  mMessageFolder->GetCanDeleteMessages(&canDelete);
  PRBool canFile = PR_FALSE;
  mMessageFolder->GetCanFileMessages(&canFile);
  if (!canDelete || !canFile)
    return NS_ERROR_FAILURE;

  // create an output stream on a temporary file. This stream will save the modified
  // message data to a file which we will later use to replace the existing message.
  // The file is removed in the destructor.
  rv = GetSpecialDirectoryWithFileName(NS_OS_TEMP_DIR, "nsmail.tmp",
                                       getter_AddRefs(mMsgFile));
  NS_ENSURE_SUCCESS(rv,rv);

  // For temp file, we should use restrictive 00600 instead of ATTACHMENT_PERMISSION 
  rv = mMsgFile->CreateUnique(nsIFile::NORMAL_FILE_TYPE, 00600);
  NS_ENSURE_SUCCESS(rv,rv);

  rv = MsgNewBufferedFileOutputStream(getter_AddRefs(mMsgFileStream), mMsgFile, -1, ATTACHMENT_PERMISSION);

  // create the additional header for data conversion. This will tell the stream converter
  // which MIME emitter we want to use, and it will tell the MIME emitter which attachments
  // should be deleted.
  const char * partId;
  const char * nextField;
  nsCAutoString sHeader("attach&del=");
  nsCAutoString detachToHeader("&detachTo=");
  for (PRUint32 u = 0; u < mAttach->mCount; ++u)
  {
    if (u > 0)
    {
      sHeader.Append(",");
      if (detaching)
        detachToHeader.Append(",");
    }
    partId = GetAttachmentPartId(mAttach->mAttachmentArray[u].mUrl);
    nextField = PL_strchr(partId, '&');
    sHeader.Append(partId, nextField ? nextField - partId : -1);
    if (detaching)
      detachToHeader.Append(mDetachedFileUris.CStringAt(u)->get());
  }

  if (detaching)
    sHeader.Append(detachToHeader);
  // stream this message to our listener converting it via the attachment mime
  // converter. The listener will just write the converted message straight to disk.
  nsCOMPtr<nsISupports> listenerSupports;
  rv = this->QueryInterface(NS_GET_IID(nsISupports), getter_AddRefs(listenerSupports));
  NS_ENSURE_SUCCESS(rv,rv);
  nsCOMPtr<nsIUrlListener> listenerUrlListener = do_QueryInterface(listenerSupports, &rv);
  NS_ENSURE_SUCCESS(rv,rv);
  rv = mMessageService->StreamMessage(
    messageUri,           // aMessageURI
    listenerSupports,     // aConsumer
    mMsgWindow,           // aMsgWindow
    listenerUrlListener,  // aUrlListener
    PR_TRUE,              // aConvertData
    sHeader.get(),        // aAdditionalHeader
    nsnull );             // requestUri
  NS_ENSURE_SUCCESS(rv,rv);

  return NS_OK;
}

// ------------------------------------

NS_IMETHODIMP
nsMessenger::DetachAttachment(const char * aContentType, const char * aUrl,
                              const char * aDisplayName, const char * aMessageUri,
                              PRBool aSaveFirst)
{
  NS_ENSURE_ARG_POINTER(aContentType);
  NS_ENSURE_ARG_POINTER(aUrl);
  NS_ENSURE_ARG_POINTER(aDisplayName);
  NS_ENSURE_ARG_POINTER(aMessageUri);

  // convenience function for JS, processing handled by DetachAllAttachments()
  return DetachAllAttachments(1, &aContentType, &aUrl, &aDisplayName, &aMessageUri, aSaveFirst);
}

NS_IMETHODIMP
nsMessenger::DetachAllAttachments(PRUint32 aCount,
                                  const char ** aContentTypeArray,
                                  const char ** aUrlArray,
                                  const char ** aDisplayNameArray,
                                  const char ** aMessageUriArray,
                                  PRBool aSaveFirst)
{
  NS_ENSURE_ARG_MIN(aCount, 1);
  NS_ENSURE_ARG_POINTER(aContentTypeArray);
  NS_ENSURE_ARG_POINTER(aUrlArray);
  NS_ENSURE_ARG_POINTER(aDisplayNameArray);
  NS_ENSURE_ARG_POINTER(aMessageUriArray);

  if (aSaveFirst)
    return SaveAllAttachments(aCount, aContentTypeArray, aUrlArray, aDisplayNameArray, aMessageUriArray, PR_TRUE);
  else
    return DetachAttachments(aCount, aContentTypeArray, aUrlArray, aDisplayNameArray, aMessageUriArray, nsnull);
}

nsresult
nsMessenger::DetachAttachments(PRUint32 aCount,
                                  const char ** aContentTypeArray,
                                  const char ** aUrlArray,
                                  const char ** aDisplayNameArray,
                                  const char ** aMessageUriArray,
                                  nsCStringArray *saveFileUris,
                                  PRBool withoutWarning)
{
  // if withoutWarning no dialog for user
  if (!withoutWarning && NS_FAILED(PromptIfDeleteAttachments(saveFileUris != nsnull, aCount, aDisplayNameArray)))
      return NS_OK;

  nsresult rv = NS_OK;


  // ensure that our arguments are valid
//  char * partId;
  for (PRUint32 u = 0; u < aCount; ++u)
  {
    // ensure all of the message URI are the same, we cannot process
    // attachments from different messages
    if (u > 0 && 0 != strcmp(aMessageUriArray[0], aMessageUriArray[u]))
    {
      rv = NS_ERROR_INVALID_ARG;
      break;
    }

    // ensure that we don't have deleted messages in this list
    if (0 == strcmp(aContentTypeArray[u], MIMETYPE_DELETED))
    {
      rv = NS_ERROR_INVALID_ARG;
      break;
    }

    // for the moment we prevent any attachments other than root level
    // attachments being deleted (i.e. you can't delete attachments from a
    // email forwarded as an attachment). We do this by ensuring that the
    // part id only has a single period in it (e.g. "1.2").
    //TODO: support non-root level attachment delete
//    partId = ::GetAttachmentPartId(aUrlArray[u]);
//    if (!partId || PL_strchr(partId, '.') != PL_strrchr(partId, '.'))
//    {
//      rv = NS_ERROR_INVALID_ARG;
//      break;
//    }
  }
  if (NS_FAILED(rv))
  {
    Alert("deleteAttachmentFailure");
    return rv;
  }

  //TODO: ensure that nothing else is processing this message uri at the same time

  //TODO: if any of the selected attachments are messages that contain other
  // attachments we need to warn the user that all sub-attachments of those
  // messages will also be deleted. Best to display a list of them.

  // get the listener for running the url
  nsDelAttachListener * listener = new nsDelAttachListener;
  if (!listener)
    return NS_ERROR_OUT_OF_MEMORY;
  nsCOMPtr<nsISupports> listenerSupports; // auto-delete of the listener with error
  listener->QueryInterface(NS_GET_IID(nsISupports), getter_AddRefs(listenerSupports));

  if (saveFileUris)
    listener->mDetachedFileUris = *saveFileUris;
  // create the attachments for use by the listener
  nsAttachmentState * attach = new nsAttachmentState;
  if (!attach)
    return NS_ERROR_OUT_OF_MEMORY;
  rv = attach->Init(aCount, aContentTypeArray, aUrlArray, aDisplayNameArray, aMessageUriArray);
  if (NS_SUCCEEDED(rv))
    rv = attach->PrepareForAttachmentDelete();
  if (NS_FAILED(rv))
  {
    delete attach;
    return rv;
  }

  // initialize our listener with the attachments and details. The listener takes ownership
  // of 'attach' immediately irrespective of the return value (error or not).
  return listener->StartProcessing(this, mMsgWindow, attach, saveFileUris != nsnull);
}

nsresult
nsMessenger::PromptIfDeleteAttachments(PRBool aSaveFirst,
                                       PRUint32 aCount,
                                       const char ** aDisplayNameArray)
{
  nsresult rv = NS_ERROR_FAILURE;

  nsCOMPtr<nsIPrompt> dialog(do_GetInterface(mDocShell));
  if (!dialog) return rv;

  if (!mStringBundle)
  {
    rv = InitStringBundle();
    NS_ENSURE_SUCCESS(rv, rv);
  }

  // create the list of attachments we are removing
  nsString displayString;
  nsString attachmentList;
  for (PRUint32 u = 0; u < aCount; ++u)
  {
    rv = ConvertAndSanitizeFileName(aDisplayNameArray[u], getter_Copies(displayString), nsnull);
    NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
    attachmentList.Append(displayString);
    attachmentList.Append(PRUnichar('\n'));
  }
  const PRUnichar *formatStrings[] = { attachmentList.get() };

  // format the message and display
  nsString promptMessage;
  const PRUnichar * propertyName = aSaveFirst ?
    NS_LITERAL_STRING("detachAttachments").get() : NS_LITERAL_STRING("deleteAttachments").get();
  rv = mStringBundle->FormatStringFromName(propertyName, formatStrings, 1,getter_Copies(promptMessage));
  NS_ENSURE_SUCCESS(rv, rv);

  PRBool dialogResult = PR_FALSE;
  rv = dialog->Confirm(nsnull, promptMessage.get(), &dialogResult);
  NS_ENSURE_SUCCESS(rv, rv);

  return dialogResult ? NS_OK : NS_ERROR_FAILURE;
}

/**
 * converts an html mail snippet to plain text
 * @param aMessageText - in place conversion
 */
nsresult nsMessenger::convertMsgSnippetToPlainText(nsAString& aMessageText)
{
  nsString bodyText;
  nsresult rv = NS_OK;

  // Create a parser
  nsCOMPtr<nsIParser> parser = do_CreateInstance(kCParserCID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  // Create the appropriate output sink
  nsCOMPtr<nsIContentSink> sink = do_CreateInstance(NS_PLAINTEXTSINK_CONTRACTID,&rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIHTMLToTextSink> textSink(do_QueryInterface(sink));
  NS_ENSURE_TRUE(textSink, NS_ERROR_FAILURE);
  // include nsIDocumentEncoder::OutputFormatted so that we don't miss URLs on background indexing
  // see bug #3138 for details
  PRUint32 flags = nsIDocumentEncoder::OutputLFLineBreak
                   | nsIDocumentEncoder::OutputNoScriptContent
                   | nsIDocumentEncoder::OutputNoFramesContent
                   | nsIDocumentEncoder::OutputBodyOnly
                   | nsIDocumentEncoder::OutputFormatted;
  // if you change the wrap width here (80 characters) be sure to make a corresponding change to the hr delimeter
  // string in pbDataPane.js.
  textSink->Initialize(&bodyText, flags, 80);
  parser->SetContentSink(sink);
  rv = parser->Parse(aMessageText, 0, NS_LITERAL_CSTRING("text/html"), PR_TRUE);

  aMessageText.Assign(bodyText);
  return rv;
}

/**
 * stripQuotesFromMsgSnippet - Reduces quoted reply text including the citation (Scott wrote:) from
 *                             the message snippet to " ... ". Assumes the snippet has been decoded and converted to
 *                             plain text.
 * @param aMsgSnippet        - The string to strip quotes from.
 * @param aCompressedQuotes  - The resulting string with quotes stripped.
 */
void nsMessenger::compressQuotesInMsgSnippet(const nsAString& aMsgSnippet, nsAString& aCompressedQuotes)
{
  PRUint32 msgBodyStrLen = aMsgSnippet.Length();
  PRBool lastLineWasAQuote = PR_FALSE;
  PRUint32 offset = 0;
  PRUint32 lineFeedPos = 0;

  while (offset < msgBodyStrLen)
  {
    lineFeedPos = aMsgSnippet.FindChar('\n', offset);
    if (lineFeedPos != kNotFound)
    {
      const nsAString& currentLine = Substring(aMsgSnippet, offset, lineFeedPos - offset);
      // this catches quoted text ("> "), nested quotes of any level (">> ", ">>> ", ...)
      // it also catches empty line quoted text (">"). 
      // Try to strip the citation. If the current line ends with a ':' and the next line
      // looks like a quoted reply (starts with a ">") skip the current line
      // it'd be nice if we could also detect signature text and skip that too
      if (StringBeginsWith(currentLine, NS_LITERAL_STRING(">")) ||
          (lineFeedPos + 1 < msgBodyStrLen  && lineFeedPos
          && aMsgSnippet[lineFeedPos - 1] == PRUnichar(':')
          && (aMsgSnippet[lineFeedPos + 1] == PRUnichar('>') || aMsgSnippet[lineFeedPos + 1] == PRUnichar('\n'))))
      {
        lastLineWasAQuote = PR_TRUE;
      }
      else if (!currentLine.IsEmpty())
      {
        if (lastLineWasAQuote)
        {
          aCompressedQuotes += NS_LITERAL_STRING(" ... ");
          lastLineWasAQuote = PR_FALSE;
        }

        aCompressedQuotes += currentLine;
        aCompressedQuotes += PRUnichar(' '); // don't forget to substitute a space for the line feed
      }

      offset = lineFeedPos + 1;
    }
    else
    {
      aCompressedQuotes.Append(Substring(aMsgSnippet, offset, msgBodyStrLen - offset));
      break;
    }
  }
}

static PRBool HasWordHRChild(nsIDOMNode *aNode, PRBool *aFoundHR)
{
  nsCOMPtr<nsIDOMNodeList> childNodeList;
  nsresult rv = aNode->GetChildNodes(getter_AddRefs(childNodeList));
  if (NS_SUCCEEDED(rv) && childNodeList)
  {
    PRUint32 childListLength = 0;
    (void) childNodeList->GetLength(&childListLength);

    for (PRUint32 i = 0; i < childListLength && !*aFoundHR; i++)
    {
      nsCOMPtr<nsIDOMNode> childNode;
      childNodeList->Item(i, getter_AddRefs(childNode));
      if (!childNode)
        continue; // unexpected

      nsAutoString nodeName;
      PRUint16 nodeType;
      rv = childNode->GetNodeType(&nodeType);
      NS_ENSURE_SUCCESS(rv, rv);
      rv = childNode->GetLocalName(nodeName);
      NS_ENSURE_SUCCESS(rv, rv);

      // XXX todo
      // if we get false positives, we could look for these attributes
      // <hr size="2" width="100%" align="center" tabindex="-1">
      if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("hr"))
      {
        *aFoundHR = PR_TRUE;
         return NS_OK;
      }
      else
      {
        rv = HasWordHRChild(childNode, aFoundHR);
        NS_ENSURE_SUCCESS(rv, rv);
      }
    }
  }
  return NS_OK;
}

static nsresult HasWordSection1Div(nsIDOMNode *aGrandChildNode, PRBool *aFoundDiv)
{
  nsAutoString nodeName;
  PRUint16 nodeType;
  nsresult rv = aGrandChildNode->GetNodeType(&nodeType);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = aGrandChildNode->GetLocalName(nodeName);
  NS_ENSURE_SUCCESS(rv, rv);
  
  if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("div"))
  {
    nsCOMPtr<nsIDOMNode> greatGrandChildNode;
    rv = aGrandChildNode->GetFirstChild(getter_AddRefs(greatGrandChildNode));
    while (NS_SUCCEEDED(rv) && greatGrandChildNode && !*aFoundDiv)
    {
      // look for a <div> with the special border
      // it may be preceeded by text nodes 
      nsCOMPtr<nsIDOMElement> greatGrandChildElement = do_QueryInterface(greatGrandChildNode, &rv);
      if (NS_SUCCEEDED(rv) && greatGrandChildElement)
      {
        nsAutoString styleAttributeValue;
        rv = greatGrandChildElement->GetAttribute(NS_LITERAL_STRING("style"), styleAttributeValue);
        NS_ENSURE_SUCCESS(rv, rv);
#ifdef DEBUG
        printf("XYZ style [%s]\n", NS_LossyConvertUTF16toASCII(styleAttributeValue).get());
#endif
        if (styleAttributeValue.EqualsLiteral("border-right: medium none; border-width: 1pt medium medium; border-style: solid none none; border-color: rgb(181, 196, 223) -moz-use-text-color -moz-use-text-color; padding: 3pt 0in 0in;"))
        {
          *aFoundDiv = PR_TRUE;
        }
      }
      rv = greatGrandChildNode->GetNextSibling(getter_AddRefs(greatGrandChildNode));
    }

    // handle the Outlook scenario:
    //
    // <div class="Section1">
    // <p class="MsoNormal"></p>
    // <p class="MsoNormal"></p>
    // <p class="MsoNormal"></p>
    // <div>
    //   <div class="MsoNormal">
    //     <font>
    //       <span>
    //         #text
    //         <hr size="2" width="100%" align="center" tabindex="-1">
    //         #text
    //      </span>
    //      </font>
    //   <div>
    // </div>
    // <p class="MsoNormal"></p>
    //
    // see bug #6748 for details
    if (!*aFoundDiv)
    {
      rv = HasWordHRChild(aGrandChildNode, aFoundDiv);
      NS_ENSURE_SUCCESS(rv, rv);
    }
  }

  if (*aFoundDiv)
  {
    nsCOMPtr <nsIDOM3Node> dom3Node = do_QueryInterface(aGrandChildNode, &rv);
    if (NS_SUCCEEDED(rv) && dom3Node)
    {
      nsAutoString textContent;
      rv = dom3Node->GetTextContent(textContent);
      NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG
      textContent.Trim(" ");
      textContent.StripChars("\r\n");
      printf("XYZ textContent [%s]\n", NS_LossyConvertUTF16toASCII(textContent).get());

      // XXX todo
      // to avoid false positives, should we set *aFoundDiv to PR_FALSE if textContent does not being with "From:"?
#endif
    }
  }
  return NS_OK;
}

static PRBool IsFromOrDateHeaderTextNode(nsIDOMNode *aNode)
{
  nsAutoString nodeValue;
  nsresult rv = aNode->GetNodeValue(nodeValue);
  if (NS_SUCCEEDED(rv))
  {
    // our text node might have some leading whitespace
    nodeValue.StripChars("\r\n ");
#ifdef DEBUG
    printf("XYZ looking for 'From:' or 'Date:' [%s]\n", NS_LossyConvertUTF16toASCII(nodeValue).get());
#endif
    return (nodeValue.Length() > 6 && 
            (Substring(nodeValue, 0, 5).EqualsLiteral("From:") || 
             Substring(nodeValue, 0, 5).EqualsLiteral("Date:") || 
             Substring(nodeValue, 0, 4).EqualsLiteral("Von:") ||
             Substring(nodeValue, 0, 6).EqualsLiteral("Datum:")));
  }
  return PR_FALSE;
}

static PRBool IsOriginalMessageTextNode(nsIDOMNode *aNode)
{
  nsAutoString nodeValue;
  nsresult rv = aNode->GetNodeValue(nodeValue);
  if (NS_SUCCEEDED(rv))
  {
    nodeValue.StripChars("\r\n -");
#ifdef DEBUG
    printf("XYZ IsOriginalMessageTextNode [%s]\n", NS_LossyConvertUTF16toASCII(nodeValue).get());
#endif
    return (nodeValue.EqualsLiteral("OriginalMessage") ||
            nodeValue.EqualsLiteral("OriginalNachricht"));
  }
  return PR_FALSE;
}
          
static nsresult RemoveQuoteNodes(nsIDOMNode *outNode)
{
  nsAutoString nodeName;
  PRUint16 nodeType;
  nsresult rv = outNode->GetNodeType(&nodeType);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = outNode->GetLocalName(nodeName);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIDOMNodeList> nodeList;
  rv = outNode->GetChildNodes(getter_AddRefs(nodeList));
  if (NS_SUCCEEDED(rv) && nodeList) 
  { 
    PRUint32 listLength = 0; 
    (void) nodeList->GetLength(&listLength); 
    if (listLength == 0)
    {
      return NS_OK;
    }
    
    nsCOMPtr<nsIDOMNode> childNode;
    for (PRUint32 i = listLength; i >= 1; i--) 
    { 
      nodeList->Item(i-1, getter_AddRefs(childNode));
      if (!childNode)
        continue; // unexpected
        
      nsAutoString nodeName;
      PRUint16 nodeType;
      rv = childNode->GetNodeType(&nodeType);
      NS_ENSURE_SUCCESS(rv, rv);
      rv = childNode->GetLocalName(nodeName);
      NS_ENSURE_SUCCESS(rv, rv);
      PRBool removeElement = PR_FALSE;
      PRBool isReplbqBlockquote = PR_FALSE;
      PRBool isReplbqDiv = PR_FALSE;
      PRBool isGmailQuoteDiv = PR_FALSE; 
      PRBool isWordSection1Div = PR_FALSE;
      PRBool isHotmailHr = PR_FALSE;
      PRBool isOutlookOriginalMessage = PR_FALSE;

      if (nodeType == nsIDOMNode::TEXT_NODE && IsOriginalMessageTextNode(childNode))
      {
        nsCOMPtr<nsIDOMNode> nextSibling;
        rv = childNode->GetNextSibling(getter_AddRefs(nextSibling));
        if (NS_SUCCEEDED(rv) && nextSibling)
        {
          PRUint16 nextNodeType;
          rv = nextSibling->GetNodeType(&nextNodeType);
          NS_ENSURE_SUCCESS(rv, rv);
 
          nsAutoString nextNodeName;
          rv = nextSibling->GetLocalName(nextNodeName);
          NS_ENSURE_SUCCESS(rv, rv);
 
          if (nextNodeType == nsIDOMNode::ELEMENT_NODE && nextNodeName.EqualsLiteral("br"))
          {
            nsCOMPtr<nsIDOMNode> nextNextSibling;
            rv = nextSibling->GetNextSibling(getter_AddRefs(nextNextSibling));
            if (NS_SUCCEEDED(rv) && nextNextSibling)
            {
              PRUint16 nextNextNodeType;
              rv = nextNextSibling->GetNodeType(&nextNextNodeType);
              NS_ENSURE_SUCCESS(rv, rv);
              
              if (nextNextNodeType == nsIDOMNode::TEXT_NODE && IsFromOrDateHeaderTextNode(nextNextSibling))
              {
                removeElement = PR_TRUE;
                isOutlookOriginalMessage = PR_TRUE;
              }
            }
          }
        }
      }
      else if (nodeType == nsIDOMNode::ELEMENT_NODE)
      {
        if (nodeName.EqualsLiteral("blockquote"))
        {
          removeElement = PR_TRUE;

          nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(childNode, &rv);
          NS_ENSURE_SUCCESS(rv, rv);

          nsAutoString className;
          rv = domHTMLElement->GetClassName(className);
          NS_ENSURE_SUCCESS(rv, rv);

          if (className.EqualsLiteral("replbq"))
          {
            isReplbqBlockquote = PR_TRUE;
          }
        }
        else if (nodeName.EqualsLiteral("div"))
        {
          nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(childNode, &rv);
          NS_ENSURE_SUCCESS(rv, rv);

          nsAutoString className;
          rv = domHTMLElement->GetClassName(className);
          NS_ENSURE_SUCCESS(rv, rv);

          if (className.EqualsLiteral("gmail_quote"))
          {
            removeElement = PR_TRUE;
            isGmailQuoteDiv = PR_TRUE;
          }
          else if (className.EqualsLiteral("replbq"))
          {
            removeElement = PR_TRUE;
            isReplbqDiv = PR_TRUE;
          }
          else if (className.EqualsLiteral("WordSection1") || className.EqualsLiteral("Section1"))
          {
            // see bug #6748 for details
            isWordSection1Div = PR_TRUE;
          }
        }
        else if (nodeName.EqualsLiteral("pre"))
        {
          nsCOMPtr<nsIDOMNSHTMLElement> domNSElement = do_QueryInterface(childNode, &rv);
          NS_ENSURE_SUCCESS(rv, rv);

          nsAutoString innerHTML;
          rv = domNSElement->GetInnerHTML(innerHTML);
          NS_ENSURE_SUCCESS(rv, rv);
 
          if (!innerHTML.IsEmpty())
          {
            // going from <pre> to <div>, need newlines to be <br>s
            innerHTML.ReplaceSubstring(NS_LITERAL_STRING("\n"), NS_LITERAL_STRING("<br />"));
              
            nsCOMPtr<nsIDOMDocument> ownerDocument;
            rv = outNode->GetOwnerDocument(getter_AddRefs(ownerDocument));
            NS_ENSURE_SUCCESS(rv, rv);

            // set the document uri before we call SetInnerHTML()
            // see bug #7151 for details
            nsCOMPtr<nsIDocument> doc = do_QueryInterface(ownerDocument, &rv);
            NS_ENSURE_SUCCESS(rv, rv);

            nsCOMPtr<nsIURI> uri;
            NS_NewURI(getter_AddRefs(uri), NS_LITERAL_CSTRING("about:blank"));
            NS_ENSURE_TRUE(uri, NS_ERROR_UNEXPECTED);
            doc->SetDocumentURI(uri);
            
            nsCOMPtr<nsIDOMElement> divElement;
            rv = ownerDocument->CreateElement(NS_LITERAL_STRING("div"), getter_AddRefs(divElement));
            NS_ENSURE_SUCCESS(rv, rv);

            nsCOMPtr<nsIDOMNSHTMLElement> divNSElement = do_QueryInterface(divElement, &rv);
            NS_ENSURE_SUCCESS(rv, rv);
              
            rv = divNSElement->SetInnerHTML(innerHTML);
            NS_ENSURE_SUCCESS(rv, rv);
              
            nsCOMPtr<nsIDOMNode> divNode = do_QueryInterface(divElement, &rv);
            NS_ENSURE_SUCCESS(rv, rv);

            nsCOMPtr<nsIDOMNode> tmp;
            rv = outNode->ReplaceChild(divNode, childNode, getter_AddRefs(tmp));
            NS_ENSURE_SUCCESS(rv, rv);
          }
        }
        else if (nodeName.EqualsLiteral("hr"))
        {
          nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(childNode, &rv);
          if (NS_SUCCEEDED(rv) && domHTMLElement)
          {
            nsAutoString id;
            rv = domHTMLElement->GetId(id);
            if (NS_SUCCEEDED(rv) && id.EqualsLiteral("stopSpelling"))
            {
              removeElement = PR_TRUE;
              isHotmailHr = PR_TRUE;
            }
          }
        }
        else
        {
          // leave it alone
        }
      }

      if (isHotmailHr || isOutlookOriginalMessage)
      {
        // handle the following Hotmail scenario:
        // <div>
        //   #text
        //   <br>
        //   <br>
        //   <hr id="stopSpelling">
        //   ...
        // </div>
        //
        // remove all siblings after the <hr> that contains the special id
        // but stop at the <hr class="__pbConvHr"> node that we emitted in msgDisplayEmitter.js
        // see bug #6763 for details
        //
        // or handle the following Outlook/Zimbra scenarios:
        //
        // <p>
        //   <br>
        //   #text (----- Original Message -----) 
        //   <br>
        //   #text (From:)
        //   ...
        // </p>
        //
        // and
        //
        // <div>
        //  <br>
        //  #text (----- Original Message -----) 
        //  <br>
        //  #text (From:)
        //  ...
        // </div>
        //
        // remove all siblings after the text node that contains "----- Original Message -----" (or "-----Original Message-----")
        // but stop at the <hr class="__pbConvHr"> node that we emitted in msgDisplayEmitter.js
        // see bug #6751 for details
        nsCOMPtr<nsIDOMNode> nextSibling;
        rv = childNode->GetNextSibling(getter_AddRefs(nextSibling));
        while (NS_SUCCEEDED(rv) && nextSibling)
        {
          nsAutoString className;
          nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(nextSibling);
          if (domHTMLElement && NS_SUCCEEDED(domHTMLElement->GetClassName(className)) && className.EqualsLiteral("__pbConvHr"))
          {
            nextSibling = nsnull;
          }
          else
          {
            nsCOMPtr<nsIDOMNode> tmp;
            rv = outNode->RemoveChild(nextSibling, getter_AddRefs(tmp));
            NS_ENSURE_SUCCESS(rv, rv);
            rv = childNode->GetNextSibling(getter_AddRefs(nextSibling));
          }
        }

        // we'll remove the <hr> down below
      }
      else if (isWordSection1Div)
      {
        // handle the following Outlook scenarios:
        // 
        // <div class="WordSection1"> (or <div class="Section1">)
        //  <p class="MsoNormal"></p>
        //  <p class="MsoNormal"></p>
        //  <p class="MsoNormal"></p>
        //  <div>
        //    <div style="border-right: medium none; border-width: 1pt medium medium; border-style: solid none none; border-color: rgb(181, 196, 223) -moz-use-text-color -moz-use-text-color; padding: 3pt 0in 0in;">
        //    </div>
        //  </div>
        //  <p class="MsoNormal"></p>
        // </div>
        //
        // and
        //
        // <div class="Section1">
        //  <p class="MsoNormal"></p>
        //  <p class="MsoNormal"></p>
        //  <p class="MsoNormal"></p>
        //  <div>
        //    #textNode
        //    <div style="border-right: medium none; border-width: 1pt medium medium; border-style: solid none none; border-color: rgb(181, 196, 223) -moz-use-text-color -moz-use-text-color; padding: 3pt 0in 0in;">
        //    </div>
        //  </div>
        //  <p class="MsoNormal"></p>
        // </div>
        //
        // see bug #6748 for details
        nsCOMPtr<nsIDOMNodeList> childNodeList;
        rv = childNode->GetChildNodes(getter_AddRefs(childNodeList));
        if (NS_SUCCEEDED(rv) && childNodeList)
        {
          PRUint32 childListLength = 0;
          (void) childNodeList->GetLength(&childListLength);

          PRBool foundDiv = PR_FALSE;
          for (PRUint32 i = 0; i < childListLength && !foundDiv; i++)
          {
            nsCOMPtr<nsIDOMNode> grandChildNode;
            childNodeList->Item(i, getter_AddRefs(grandChildNode));
            if (!grandChildNode)
              continue; // unexpected
  
            rv = HasWordSection1Div(grandChildNode, &foundDiv);
            if (NS_SUCCEEDED(rv) && foundDiv)
            {
              // remove all siblings after the <div> that contains the <div> with the special border
              nsCOMPtr<nsIDOMNode> nextSibling;
              rv = grandChildNode->GetNextSibling(getter_AddRefs(nextSibling));
              while (NS_SUCCEEDED(rv) && nextSibling)
              {
                nsCOMPtr<nsIDOMNode> tmp;
                rv = childNode->RemoveChild(nextSibling, getter_AddRefs(tmp));
                NS_ENSURE_SUCCESS(rv, rv);
                rv = grandChildNode->GetNextSibling(getter_AddRefs(nextSibling));
              }
            
              // now remove the <div> that contains the <div> with the special border
              nsCOMPtr<nsIDOMNode> tmp;
              rv = childNode->RemoveChild(grandChildNode, getter_AddRefs(tmp));
              NS_ENSURE_SUCCESS(rv, rv);
            }
          }
        }
      }
      else if (removeElement && isGmailQuoteDiv)
      {
        nsCOMPtr<nsIDOMNodeList> childNodeList;
        rv = childNode->GetChildNodes(getter_AddRefs(childNodeList));
        if (NS_SUCCEEDED(rv) && childNodeList)
        {
          PRUint32 childListLength = 0;
          (void) childNodeList->GetLength(&childListLength);
          
          if (childListLength == 0)
          {
            return NS_OK;
          }

          PRBool divAfterBlockquote = PR_FALSE;        
          nsCOMPtr<nsIDOMNode> gmailQuoteBlockQuoteNode;
          for (PRUint32 i = childListLength; i >= 1; i--)
          {
            nsCOMPtr<nsIDOMNode> grandChildNode;
            childNodeList->Item(i-1, getter_AddRefs(grandChildNode));
            if (!grandChildNode)
              continue; // unexpected
  
            nsAutoString nodeName;
            PRUint16 nodeType;
            rv = grandChildNode->GetNodeType(&nodeType);
            NS_ENSURE_SUCCESS(rv, rv);
            rv = grandChildNode->GetLocalName(nodeName);
            NS_ENSURE_SUCCESS(rv, rv);
            
            // handle this gmail "reply below quote" scenario:
            // <div class=gmail_quote>
            //   Seth Spitzer <span>&lt;<a href="mailto:sspitzer@gmail.com">sspitzer@gmail.com</a>&gt;</span> wrote:
            //   <br>
            //   <blockquote class=gmail_quote><div>x y z</div></blockquote>
            //   <div>a b c</div>
            // </div>
            // see bug #6738 for details
            //
            // but watch out for this gmail "reply above quote" scenario:
            // <div>a b c</div> 
            // <div class=gmail_quote>
            //   Seth Spitzer <span>&lt;<a href="mailto:sspitzer@gmail.com">sspitzer@gmail.com</a>&gt;</span> wrote:
            //   <br>
            //   <blockquote class=gmail_quote><div>x y z</div></blockquote>
            // </div>
            // see bug #6766 for details
            if (gmailQuoteBlockQuoteNode)
            {
              if (nodeType == nsIDOMNode::TEXT_NODE)
              {
                nsAutoString grandChildNodeValue;
                rv = grandChildNode->GetNodeValue(grandChildNodeValue);
                NS_ENSURE_SUCCESS(rv, rv);
#ifdef DEBUG
                printf("XYZ remove [%s]\n", NS_LossyConvertUTF16toASCII(grandChildNodeValue).get());
#endif
                rv = grandChildNode->SetNodeValue(NS_LITERAL_STRING(""));
                NS_ENSURE_SUCCESS(rv, rv);
              }
              else if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("br"))
              {
#ifdef DEBUG
                printf("XYZ remove the <br>\n");
#endif
                nsCOMPtr<nsIDOMNode> tmp;
                rv = childNode->RemoveChild(grandChildNode, getter_AddRefs(tmp));
                NS_ENSURE_SUCCESS(rv, rv);
              }
              else if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("span"))
              {
                nsCOMPtr<nsIDOMNSHTMLElement> domNSElement = do_QueryInterface(grandChildNode, &rv);
                NS_ENSURE_SUCCESS(rv, rv);

                nsAutoString innerHTML;
                rv = domNSElement->GetInnerHTML(innerHTML);
                NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG
                printf("XYZ remove [%s]\n", NS_LossyConvertUTF16toASCII(innerHTML).get());
#endif
                
                nsCOMPtr<nsIDOMDocument> ownerDocument;
                rv = outNode->GetOwnerDocument(getter_AddRefs(ownerDocument));
                NS_ENSURE_SUCCESS(rv, rv);
                
                // set the document uri before we call SetInnerHTML()
                // see bug #7151 for details
                nsCOMPtr<nsIDocument> doc = do_QueryInterface(ownerDocument, &rv);
                NS_ENSURE_SUCCESS(rv, rv);
                
                nsCOMPtr<nsIURI> uri;
                NS_NewURI(getter_AddRefs(uri), NS_LITERAL_CSTRING("about:blank"));
                NS_ENSURE_TRUE(uri, NS_ERROR_UNEXPECTED);
                doc->SetDocumentURI(uri);
                
                rv = domNSElement->SetInnerHTML(NS_LITERAL_STRING(""));
                NS_ENSURE_SUCCESS(rv, rv);
              }  
            }
            else if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("div"))
            {
              // there is a div after the blockquote, so this is gmail "reply after quote"
              // see bug #6738 for details
              divAfterBlockquote = PR_TRUE;
            }
            else if ((nodeType == nsIDOMNode::ELEMENT_NODE) && nodeName.EqualsLiteral("blockquote"))
            {
              nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(grandChildNode, &rv);
              NS_ENSURE_SUCCESS(rv, rv);

              nsAutoString className;
              rv = domHTMLElement->GetClassName(className);
              NS_ENSURE_SUCCESS(rv, rv);

              if (className.EqualsLiteral("gmail_quote"))
              {
                gmailQuoteBlockQuoteNode = grandChildNode;
              }
            } 
          }
          
          // if there is no div after the blockquote, so this is gmail "reply before quote"
          // so we we we want to remove the outer div, see bug #6766 for details
          if (gmailQuoteBlockQuoteNode && divAfterBlockquote)
          { 
            childNode = gmailQuoteBlockQuoteNode;
            removeElement = PR_FALSE;
          }
        }
      }

      if (removeElement)
      {
        // try to remove the "Scott wrote:\n" text node before the quote
        // but don't do this for the hotmail <hr>, as there is no "Scott wrote:\n" before it.
        nsCOMPtr<nsIDOMNode> previousSibling;
        rv = childNode->GetPreviousSibling(getter_AddRefs(previousSibling));
        if (NS_SUCCEEDED(rv) && previousSibling && !isHotmailHr)
        {
          PRUint16 previousSiblingNodeType;
          rv = previousSibling->GetNodeType(&previousSiblingNodeType);
          NS_ENSURE_SUCCESS(rv, rv);
  
          PRBool seenAppleInterchangeNewlineBR = PR_FALSE;
          
          // if the previous sibling is a <br> element, look back one more for the text node.
          // XXX TODO fix this to handle multiple <br>s, looking for the first text node?
          if (previousSiblingNodeType == nsIDOMNode::ELEMENT_NODE)
          {
            nsAutoString previousSiblingNodeName;
            rv = previousSibling->GetLocalName(previousSiblingNodeName);
            NS_ENSURE_SUCCESS(rv, rv);

            if (previousSiblingNodeName.EqualsLiteral("br"))
            {
              nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(previousSibling, &rv);
              NS_ENSURE_SUCCESS(rv, rv);
  
              nsAutoString className;
              rv = domHTMLElement->GetClassName(className);
              NS_ENSURE_SUCCESS(rv, rv);

              seenAppleInterchangeNewlineBR = className.EqualsLiteral("Apple-interchange-newline");
            
              rv = previousSibling->GetPreviousSibling(getter_AddRefs(previousSibling));
              if (NS_SUCCEEDED(rv) && previousSibling)
              {
                rv = previousSibling->GetNodeType(&previousSiblingNodeType);
                NS_ENSURE_SUCCESS(rv, rv);
              }              
            }
            else if (previousSiblingNodeName.EqualsLiteral("pre"))
            {
              // handle this scenario:
              // <pre>
              // On 26 Oct 2010 at 10:21:11, Scott MacGregor arranged the bits on my disk to say:
              // </pre>
              // <blockquote>
              // ...
              // </blockquote>
              //
              // but also handle this scenario:
              // <pre>
              // Seth,
              // Sound good.
              // -Scott
              //
              // On 26 Oct 2010 at 10:21:11, Scott MacGregor arranged the bits on my disk to say:
              // </pre>
              // <blockquote>
              // ...
              // </blockquote>
              //
              rv = previousSibling->GetFirstChild(getter_AddRefs(previousSibling));
              if (NS_SUCCEEDED(rv) && previousSibling)
              {
                rv = previousSibling->GetNodeType(&previousSiblingNodeType);
                NS_ENSURE_SUCCESS(rv, rv);
              }
            }
          }

          if (seenAppleInterchangeNewlineBR && previousSibling && previousSiblingNodeType == nsIDOMNode::ELEMENT_NODE)
          {
            // handle the Mail.app scenario:
            // <div>On Nov 9, 2010, at 4:22 PM, Seth Spitzer wrote:</div><br class="Apple-interchange-newline"><blockquote type="cite">...</blockquote>
            // see bug #6651 for details
            nsAutoString previousSiblingNodeName;
            rv = previousSibling->GetLocalName(previousSiblingNodeName);
            NS_ENSURE_SUCCESS(rv, rv);
            
            if (previousSiblingNodeName.EqualsLiteral("div"))
            {
              nsCOMPtr<nsIDOMNSHTMLElement> domNSElement = do_QueryInterface(previousSibling, &rv);
              NS_ENSURE_SUCCESS(rv, rv);

              nsAutoString innerHTML;
              rv = domNSElement->GetInnerHTML(innerHTML);
              NS_ENSURE_SUCCESS(rv, rv);

              PRInt32 colonPos = innerHTML.RFind(NS_LITERAL_STRING(":"));
              if (colonPos != -1)
              {
#ifdef DEBUG
                printf("XYZ remove [%s] and the <br>\n", NS_LossyConvertUTF16toASCII(innerHTML).get());
#endif
                nsCOMPtr<nsIDOMDocument> ownerDocument;
                rv = outNode->GetOwnerDocument(getter_AddRefs(ownerDocument));
                NS_ENSURE_SUCCESS(rv, rv);
                
                // set the document uri before we call SetInnerHTML()
                // see bug #7151 for details
                nsCOMPtr<nsIDocument> doc = do_QueryInterface(ownerDocument, &rv);
                NS_ENSURE_SUCCESS(rv, rv);
                
                nsCOMPtr<nsIURI> uri;
                NS_NewURI(getter_AddRefs(uri), NS_LITERAL_CSTRING("about:blank"));
                NS_ENSURE_TRUE(uri, NS_ERROR_UNEXPECTED);
                doc->SetDocumentURI(uri);

                rv = domNSElement->SetInnerHTML(NS_LITERAL_STRING(""));
                NS_ENSURE_SUCCESS(rv, rv);

                nsCOMPtr<nsIDOMNode> tmp;
                rv = outNode->RemoveChild(previousSibling, getter_AddRefs(tmp));
                NS_ENSURE_SUCCESS(rv, rv);
              }
            }
          }
          else if (previousSibling && previousSiblingNodeType == nsIDOMNode::TEXT_NODE)
          {
            nsAutoString previousSiblingNodeValue;
            rv = previousSibling->GetNodeValue(previousSiblingNodeValue);
            NS_ENSURE_SUCCESS(rv, rv);
           
            if (isReplbqDiv)
            {
              // handle the outlook scenario:
              // Seth wrote:<br>\n<div class="replbq">...</div>
              PRInt32 newlinePos = previousSiblingNodeValue.RFind(NS_LITERAL_STRING("\n"));
              if (newlinePos != -1)
              {
                nsCOMPtr<nsIDOMNode> previousPreviousSibling;
                rv = previousSibling->GetPreviousSibling(getter_AddRefs(previousPreviousSibling));
                if (NS_SUCCEEDED(rv) && previousPreviousSibling)
                {
                  PRUint16 previousPreviousSiblingNodeType;
                  rv = previousPreviousSibling->GetNodeType(&previousPreviousSiblingNodeType);
                  NS_ENSURE_SUCCESS(rv, rv);
                  if (previousPreviousSiblingNodeType == nsIDOMNode::ELEMENT_NODE)
                  {
                    nsAutoString previousPreviousNodeName;
                    rv = previousPreviousSibling->GetLocalName(previousPreviousNodeName);
                    NS_ENSURE_SUCCESS(rv, rv);
                    if (previousPreviousNodeName.EqualsLiteral("br"))
                    {
                      nsCOMPtr<nsIDOMNode> previousPreviousPreviousSibling;
                      rv = previousPreviousSibling->GetPreviousSibling(getter_AddRefs(previousPreviousPreviousSibling));
                      if (NS_SUCCEEDED(rv) && previousPreviousPreviousSibling)
                      {
                        PRUint16 previousPreviousPreviousSiblingNodeType;
                        rv = previousPreviousPreviousSibling->GetNodeType(&previousPreviousPreviousSiblingNodeType);
                        NS_ENSURE_SUCCESS(rv, rv);
                        if (previousPreviousPreviousSiblingNodeType == nsIDOMNode::TEXT_NODE)
                        {
                          nsAutoString previousPreviousPreviousSiblingNodeValue;
                          rv = previousPreviousPreviousSibling->GetNodeValue(previousPreviousPreviousSiblingNodeValue);
                          NS_ENSURE_SUCCESS(rv, rv);
                          
                          PRInt32 colonPos = previousPreviousPreviousSiblingNodeValue.RFind(NS_LITERAL_STRING(":"));
                          if (colonPos != -1)
                          {
#ifdef DEBUG
                            printf("XYZ remove [%s] and [%s] and the <br>\n", NS_LossyConvertUTF16toASCII(previousPreviousPreviousSiblingNodeValue).get(),
                                                                              NS_LossyConvertUTF16toASCII(previousSiblingNodeValue).get());
#endif                         
                            rv = previousPreviousPreviousSibling->SetNodeValue(NS_LITERAL_STRING(""));
                            NS_ENSURE_SUCCESS(rv, rv);

                            rv = previousSibling->SetNodeValue(NS_LITERAL_STRING(""));
                            NS_ENSURE_SUCCESS(rv, rv);

                            nsCOMPtr<nsIDOMNode> tmp;
                            rv = outNode->RemoveChild(previousPreviousSibling, getter_AddRefs(tmp));
                            NS_ENSURE_SUCCESS(rv, rv);
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
            else if (isReplbqBlockquote)
            {
              // handle the yahoo mail scenario:
              // <b><i>Seth Spitzer &lt;sspitzer@yahoo.com&gt;</i></b> wrote:<blockquote class="replbq">...</blockquote>
              PRInt32 colonPos = previousSiblingNodeValue.RFind(NS_LITERAL_STRING(":"));
              if (colonPos != -1)
              {
                nsCOMPtr<nsIDOMNode> previousPreviousSibling;
                rv = previousSibling->GetPreviousSibling(getter_AddRefs(previousPreviousSibling));
                if (NS_SUCCEEDED(rv) && previousPreviousSibling)
                {
                  PRUint16 previousPreviousSiblingNodeType;
                  rv = previousPreviousSibling->GetNodeType(&previousPreviousSiblingNodeType);
                  NS_ENSURE_SUCCESS(rv, rv);
                  if (previousPreviousSiblingNodeType == nsIDOMNode::ELEMENT_NODE)
                  {
                    nsAutoString previousPreviousNodeName;
                    rv = previousPreviousSibling->GetLocalName(previousPreviousNodeName);
                    NS_ENSURE_SUCCESS(rv, rv);
                    if (previousPreviousNodeName.EqualsLiteral("b"))
                    {
#ifdef DEBUG
                      printf("XYZ remove [%s] and the <b> before it\n", NS_LossyConvertUTF16toASCII(previousSiblingNodeValue).get());
#endif
                      rv = previousSibling->SetNodeValue(NS_LITERAL_STRING(""));
                      NS_ENSURE_SUCCESS(rv, rv);

                      nsCOMPtr<nsIDOMNode> tmp;
                      rv = outNode->RemoveChild(previousPreviousSibling, getter_AddRefs(tmp));
                      NS_ENSURE_SUCCESS(rv, rv);
                    }
                  }
                }
              }
            }
            else
            {
              // handle this scenario:
              // <pre>
              // On 26 Oct 2010 at 10:21:11, Scott MacGregor arranged the bits on my disk to say:
              // </pre>
              // <blockquote>
              // ...
              // </blockquote>
              //
              // but also handle this scenario:
              // <pre>
              // Seth,
              // Sound good.
              // -Scott
              //
              // On 26 Oct 2010 at 10:21:11, Scott MacGregor arranged the bits on my disk to say:
              // </pre>
              // <blockquote>
              // ...
              // </blockquote>
              //
              PRInt32 colonNewlinePos = previousSiblingNodeValue.RFind(NS_LITERAL_STRING(":\n"));
              if (colonNewlinePos != -1)
              {
                PRInt32 wroteStartPos = previousSiblingNodeValue.RFind(NS_LITERAL_STRING("\n"), colonNewlinePos);
#ifdef DEBUG                
                printf("XYZ %d %d [%s]\n", colonNewlinePos + 2, previousSiblingNodeValue.Length(), NS_LossyConvertUTF16toASCII(previousSiblingNodeValue).get());
#endif            
                // if are any non-whitespace characters after ":\n", this is probably not a salutation.
                PRUint32 from = colonNewlinePos + 2;
                PRUint32 end = previousSiblingNodeValue.Length();
                PRBool onlyWhiteSpaceAfterColonNewlinePos = PR_TRUE;
                while (from < end && onlyWhiteSpaceAfterColonNewlinePos)
                {
                  onlyWhiteSpaceAfterColonNewlinePos &= NS_IsAsciiWhitespace(previousSiblingNodeValue[from]);
                  from++;
                }
                                
                if (onlyWhiteSpaceAfterColonNewlinePos)
                {                  
#ifdef DEBUG
                  printf("XYZ was [%s]\n", NS_LossyConvertUTF16toASCII(previousSiblingNodeValue).get());
#endif
                  if (wroteStartPos != -1)
                  {
                    // just remove "\nOn 26 Oct 2010 at 10:21:11, Scott MacGregor arranged the bits on my disk to say:\n"
                    previousSiblingNodeValue.Cut(wroteStartPos, colonNewlinePos + 2 - wroteStartPos);
                  }
                  else
                  {
                    // just remove "J. Paul Reed wrote:\n"
                    previousSiblingNodeValue.Cut(0, previousSiblingNodeValue.Length());
                  }
#ifdef DEBUG
                  printf("XYZ is [%s]\n", NS_LossyConvertUTF16toASCII(previousSiblingNodeValue).get());
#endif 
              
                  rv = previousSibling->SetNodeValue(previousSiblingNodeValue);
                  NS_ENSURE_SUCCESS(rv, rv);
                }
              }
            }
          }
        }

        // remove the element from the DOM
        nsCOMPtr<nsIDOMNode> tmp;
        rv = outNode->RemoveChild(childNode, getter_AddRefs(tmp));
        NS_ENSURE_SUCCESS(rv, rv);
      }
      else
      {
        rv = RemoveQuoteNodes(childNode);
        NS_ENSURE_SUCCESS(rv, rv);
      }
    }
  }
  return NS_OK;
}

static nsresult RemoveSignaturesFromNode(nsIDOMNode *outNode)
{
  nsAutoString nodeName;
  PRUint16 nodeType;
  nsresult rv = outNode->GetNodeType(&nodeType);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = outNode->GetLocalName(nodeName);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIDOMNodeList> nodeList;
  rv = outNode->GetChildNodes(getter_AddRefs(nodeList));
  if (NS_SUCCEEDED(rv) && nodeList) 
  { 
    PRUint32 listLength = 0; 
    (void) nodeList->GetLength(&listLength); 
    if (listLength == 0)
    {
      return NS_OK;
    }
    
    nsCOMPtr<nsIDOMNode> childNode;
    for (PRUint32 i = listLength; i >= 1; i--) 
    { 
      nodeList->Item(i-1, getter_AddRefs(childNode));
      if (!childNode)
        continue; // unexpected
        
      nsAutoString nodeName;
      PRUint16 nodeType;
      rv = childNode->GetNodeType(&nodeType);
      NS_ENSURE_SUCCESS(rv, rv);
      rv = childNode->GetLocalName(nodeName);
      NS_ENSURE_SUCCESS(rv, rv);
      
      PRBool removeElement = PR_FALSE;

      if (nodeType == nsIDOMNode::ELEMENT_NODE)
      {
        if (nodeName.EqualsLiteral("div"))
        {
          nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(childNode, &rv);
          NS_ENSURE_SUCCESS(rv, rv);

          nsAutoString className;
          rv = domHTMLElement->GetClassName(className);
          NS_ENSURE_SUCCESS(rv, rv);

          removeElement = className.EqualsLiteral("moz-signature");
        }
      }
 
      if (removeElement)
      {
        // remove the <div class="moz-signature"> element from the DOM
        nsCOMPtr<nsIDOMNode> tmp;
        rv = outNode->RemoveChild(childNode, getter_AddRefs(tmp));
        NS_ENSURE_SUCCESS(rv, rv);
      }
      else
      {
        rv = RemoveSignaturesFromNode(childNode);
        NS_ENSURE_SUCCESS(rv, rv);
      }
    }
  }

  return NS_OK;
}

static nsresult RemoveQuotesFromNode(nsIDOMNode *aNode)
{
  nsresult rv;
  PRInt32 quoteStyle = nsIMsgGlobalIndex::kNonInlineQuotes;
  nsCOMPtr<nsIMsgGlobalIndex> globalIndex = do_GetService("@mozilla.org/msg-global-index;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = globalIndex->DetermineQuoteStyleWithNode(aNode, &quoteStyle);
  // we want to remove quotes if there are non-inline quotes (kNonInlineQuotes)
  // or if there are no quotes (kNoQuotes) for certain Outlook emails
  // see bug #6748 for details
  if (NS_SUCCEEDED(rv) && (quoteStyle != nsIMsgGlobalIndex::kInlineQuotes))
  {
    rv = RemoveQuoteNodes(aNode);
    NS_ENSURE_SUCCESS(rv, rv);
  }
  return NS_OK;
}


static nsresult RemoveQuotesInternal(const nsAString &aInHTML, nsAString &aOutHTML)
{
  nsresult rv;
  nsCOMPtr<nsIParser> parser = do_CreateInstance(kCParserCID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  NS_ENSURE_TRUE(parser, NS_ERROR_FAILURE);

  // NS_HTMLPARANOIDFRAGMENTSINK2_CONTRACTID?
  nsCOMPtr<nsIContentSink> sink = do_CreateInstance(NS_HTMLFRAGMENTSINK2_CONTRACTID);
  NS_ENSURE_TRUE(sink, NS_ERROR_FAILURE);
  nsCOMPtr<nsIFragmentContentSink> fragSink(do_QueryInterface(sink));
  NS_ENSURE_TRUE(fragSink, NS_ERROR_FAILURE);

  // create doc
  nsCOMPtr<nsIDocument> doc(do_CreateInstance(kHTMLDocumentCID));
  fragSink->SetTargetDocument(doc);

  // parse the fragment
  parser->SetContentSink(sink);
  parser->Parse(aInHTML, 0, NS_LITERAL_CSTRING("text/html"), PR_TRUE, eDTDMode_fragment);
  
  nsCOMPtr<nsIDOMDocumentFragment> contextfrag;
  rv = fragSink->GetFragment(PR_TRUE, getter_AddRefs(contextfrag));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr <nsIDOMNode> outNode = do_QueryInterface(contextfrag, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr <nsIDOMNode> bodyNode;
  rv = outNode->GetFirstChild(getter_AddRefs(bodyNode));
  NS_ENSURE_SUCCESS(rv, rv);

  PRUint16 nodeType;
  rv = bodyNode->GetNodeType(&nodeType);
  NS_ENSURE_SUCCESS(rv, rv);

  nsAutoString nodeName;
  rv = bodyNode->GetLocalName(nodeName);
  NS_ENSURE_SUCCESS(rv, rv);
 
  // the first child of the document fragment should be a <body>
  if (nodeType != nsIDOMNode::ELEMENT_NODE || !nodeName.EqualsLiteral("body"))
    return NS_ERROR_UNEXPECTED;

  nsCOMPtr<nsIDOMNodeList> nodeList;
  rv = bodyNode->GetChildNodes(getter_AddRefs(nodeList));
  if (NS_SUCCEEDED(rv) && nodeList)
  {
    PRUint32 listLength = 0;
    (void) nodeList->GetLength(&listLength);

    nsCOMPtr<nsIDOMNode> childNode;
    for (PRUint32 i = 0; i < listLength; i++)
    {
      nodeList->Item(i, getter_AddRefs(childNode));
      if (!childNode)
        continue; // unexpected

      rv = childNode->GetNodeType(&nodeType);
      NS_ENSURE_SUCCESS(rv, rv);

      rv = childNode->GetLocalName(nodeName);
      NS_ENSURE_SUCCESS(rv, rv);

      if (nodeType == nsIDOMNode::ELEMENT_NODE && nodeName.EqualsLiteral("div"))
      {
        nsCOMPtr <nsIDOMHTMLElement> domHTMLElement = do_QueryInterface(childNode, &rv);
        NS_ENSURE_SUCCESS(rv, rv);

        nsAutoString className;
        rv = domHTMLElement->GetClassName(className);
        NS_ENSURE_SUCCESS(rv, rv);
  
        // the summarized bodies are wrapped with <div class="__pbConvBody">...</div>
        // see msgDisplayEmitter.
        if (className.EqualsLiteral("__pbConvBody"))
        {
          nsCOMPtr<nsIDOMElement> domElement = do_QueryInterface(childNode, &rv);
          if (NS_SUCCEEDED(rv) && domElement)
          {
            nsAutoString removeQuotesAttributeValue;
            rv = domElement->GetAttribute(NS_LITERAL_STRING("__pbRmQuotes"), removeQuotesAttributeValue);
            if (NS_SUCCEEDED(rv) && removeQuotesAttributeValue.EqualsLiteral("true"))
            {
              // only remove quotes and signatures if we see __pbRmQuotes="true" on the div
              // see bug #6788 for details
              rv = RemoveQuotesFromNode(childNode);
              NS_ENSURE_SUCCESS(rv, rv);

              rv = RemoveSignaturesFromNode(childNode);
              NS_ENSURE_SUCCESS(rv, rv);
            }
          }
        }
      }
    }
  }

  nsCOMPtr<nsIDOMSerializer> serializer = do_CreateInstance(NS_XMLSERIALIZER_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = serializer->SerializeToString(outNode, aOutHTML);
  NS_ENSURE_SUCCESS(rv, rv);
  return NS_OK;
}

NS_IMETHODIMP nsMessenger::RemoveQuotes(const nsAString &aInHTML, nsAString &aOutHTML)
{
#ifdef DEBUG
  printf("XYZ RemoveQuotes in [%s]\n", NS_LossyConvertUTF16toASCII(aInHTML).get());
#endif
  nsresult rv = RemoveQuotesInternal(aInHTML, aOutHTML);
  if (NS_FAILED(rv))
  {
    // if we fail unexpectedly, we still want to summarize something (even if we can't properly remove quotes)
    aOutHTML.Assign(aInHTML);
  }
#ifdef DEBUG
  printf("XYZ RemoveQuotes out [%s]\n", NS_LossyConvertUTF16toASCII(aOutHTML).get());
#endif  
  return NS_OK;
}
  
NS_IMETHODIMP nsMessenger::ConvertAndCompress(PRBool aCompressQuotes, const nsAString &aInHTML, nsAString &aOutText)
{
  aOutText = aInHTML;
  // convert it to plain text (we are always given text/html)
  convertMsgSnippetToPlainText(aOutText);

  nsString resultString;
  // optionally remove quoted text from the snippet
  if (aCompressQuotes)
  {
    compressQuotesInMsgSnippet(aOutText, resultString);
  }
  else 
  {
    resultString = aOutText;
  }

  resultString.CompressWhitespace(PR_TRUE, PR_TRUE);
  aOutText = resultString;
  return NS_OK;
}


